-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-18
-- By Jonathan E. Wright
--
-- TODO BFA:
-- - add health and mana bars so we can see them without turning on personal buffs?
-- X need to figure out how to fix the unhide in combat issue
--   - instead of hide and unhide we just have to enable / disable mouse
--   - This fixes issues with hidden bars usurping mouse input, but doesn't fix issues with buttons locked out during combat.
-- - incoming damage can be used to predict when the player will die and put up a LoH or DS icon.
-- - if we have any usable trinkets they can be added to the skill or buff bars
-- X show cooldown on rotation
-- - consider selecting rotation skills less frequently than every frame so things feel more stable.
-- - Ret
--   - don't show blessings in a raid or party unless we can check if anyone in the party has our buff
--   - change inquisition timer to a bar?
--   - Should Wake of Ashes show if you will only waste 1 hopo?
--     - if Avenging Wrath / Crusade is on, yes
--     X if Wake is available and we have a choice between skills that will take us to 3 HoPo or 3+ HoPo,
--       choose the skill that takes us to exactly 3 HoPo
--   - Show Inquisition on cooldowns. 
--   X test Hammer of Wrath in rotation.
--     X BUG: Hammer of Wrath showing up when target has > 20% health.
--   X Add Shield of Vengeance to cooldowns.
--   - Add Flash of Light to cooldowns (with Selfless Healer)
--   X BUG: when Inquisition is not a selected talent, inquisition timer always shows
--   - Word of Glory support
--   - Hammer of Wrath when Crusade / Avenging Wrath is active
-- - Protection:
--   - have yet to try pro with this addon in BFA 
--   - notify when cleanse is necessary (non-raid only)
-- X BUG: when loading the addon as Prot, Buff 1 frame text shows.
-- X BUG: Wake of Ashes and Blade of Justice aren't showing in rotation 
--   - bad variable name used when added Prot range checking... Lua honey-badgers unknown variables.
-- - disable addon when in a vehicle or controlling another charcter
-- TODO WoD:
-- * instead of change skills and buffs, change to rotation and skills
--   - rotation would show current best skill and next available skill,
--     including situational skills like Rebuke and LoH.
--   - skills would show currently available skills, and skills on CD
-- X Show Hammer of Wrath on skills bar ... geez.
-- X Show Divine Storm and TV/FV on skills bar
-- X Sort skills not on CD by priority instead of cooldown
-- * Supplication!
-- X Show Divine Storm when Divine Purpose procs and Empowered Divine Storm procs
-- X Show Templar's Verdict / Final Verdict when Divine Purpose procs
--
-- TODO Mists:
-- * show hammer of wrath in skills bar
-- * show Divine Crusader in skills bar
-- * option for show global cooldowns on skills
-- X countdown timer on inquisition
-- * separate buffs window in to a buffs and skills window
-- * rebuke icon when target is casting an interruptible spell
-- TODO Cataclysm:
-- X inquisition timer
-- X dim when mounted
-- TODO:
-- * warn when Ret has Righteous Fury on in a raid or instance. Righteous Fury is 
--   useful in some PvE situations like holding aggro on Kilnmasters on Timeless Isle.
-- * if the player's target doesn't have aggro, show the icon for Hand of Reckoning
--   - this must be configurable and shouldn't default to on in instance and raids
-- * customize interface more
--   - allow icons to be clickable!
--   - allow several different setups for buttons or make the button windows movable
--     - just make a specific window for each button?  this would speed up checking
--       for which buffs are being shown.
-- * add hammer of wrath icon when target is below 20% health and HoW is not on cooldown
-- * allow user to choose the time before an icon is off cooldown for it to appear
--   - i.e. show Hammer of Wrath 0.25 seconds before the cooldown is up
--   - Cooldown Early Warning
-- X add Art of War
-- * add Divine Plea
-- * when in combat, check that the paladin has his preferred buffs such as Blessing
--   of Might, and a seal.
-- * show Flash Heal and Excorcism when Art of War is up.
--   - it's only necessary to show these if they're clickable, otherwise a single
--     art of war icon lets us know we can cast either.
--     - problem with this is that exorcism could be on cooldown
-- * option to show Turn Evil
-- * option to show Holy Wrath
-- * show Righteous Fury if ret points in Improved Righteous Fury
--   - for ret this should turn off in raids and instances 
-- * trinkets
-- * show Hand of Freedom when movement impaired.
-- * generalize... 
--   - choose a skill
--   - choose rules for when to show the skill
--     - mounted / not mounted
--     - in combat
--     - hostile target
--     - another buff is active
--     - choose early warning or not
--     - target health is below some %
-- -----------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- CircularBuffer class
-- If I could figure out the assinine requirements for loading multiple modules
-- in WoW's Lua I wouldn't need to add this at the beginning of this file.
--------------------------------------------------------------------------------

-- Meta class
-- head and tail always increase with the exception being when they reach math.maxinteger
-- rather than wrapping each time head or tail exceeds the max size of the buffer.
-- by simply increasing these and only wrapping at maxinteger we simplify some of the logic:
-- - inserts always happen at head % size
-- - empty is only when tail == head (no modulus necessary)
-- - full is when:
--   - head - tail >= size
-- - we don't need to store a count because it's easily derivable from head - tail
-- - when we reach maxInteger, we just subtract maxInteger from both head and tail, no 
--   wrapping. This works even if tail is negative because in Lua the % operator is
--   actual modulo, not just remainder and will always be positive.
local CircularBuffer = { }
CircularBuffer.__index = CircularBuffer
CircularBuffer.maxInteger = bit.lshift( 1, 31 )

-- Derived class method new
function CircularBuffer:print( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( text );
	end
end

function CircularBuffer:new( maxSize_ )
	-- disallow trailing nil parameters
	if ( maxSize_ == nil ) then
		self:print( "maxSize_ == nil!")
		return nil
	end

	local b = setmetatable( { head = 0, tail = 0, maxSize = maxSize_ }, CircularBuffer )
	return b
end

function CircularBuffer:incIndex( index )
	-- only wrap if we are at maximum integer value already	
	if ( index == self.maxInteger ) then
		index = 0
	else
		index = index + 1
	end
	return index
end

function CircularBuffer:boolToStr( value )
	if ( not value ) then
		return "false"
	else
		return "true"
	end
end

function CircularBuffer:nilToStr( value )
	if ( value == nil ) then
		return "nil"
	else
		return value
	end
end

function CircularBuffer:count()
	return self.head - self.tail
end

function CircularBuffer:isFull()
	return self:count() >= self.maxSize
end

function CircularBuffer:isEmpty()
	return self.head - self.tail == 0
end

function CircularBuffer:add( val )
	if ( self:isFull() ) then
		return false
	end

	self[self.head % self.maxSize] = val

	if ( self.head == self.maxInteger ) then
		self.head = self.head - self.maxInteger
		self.tail = self.tail - self.maxInteger
	end

	self.head = self.head + 1
	return true
end

function CircularBuffer:remove()
	if ( self:isEmpty() ) then
		return nil
	end

	local index = self.tail % self.maxSize
	local value = self[index]
	self.tail = self.tail + 1
	return value
end

-- because WoW doesn't support "require"
-- return CircularBuffer

--------------------------------------------------------------------------------


HoJ = { };

-- HoJ.ADDON_VERSION = 0.5; -- split hopo window from Templar's Verdict window
-- HoJ.ADDON_VERSION = 0.6; -- split icon window from Templar's Verdict window
-- HoJ.ADDON_VERSION = 0.81; -- major rewrite for Mists of Pandaria with low level pally fixes
-- HoJ.ADDON_VERSION = 0.82; -- ret skill bar for rebuke, blinding light, hammer of justice / fist of justice, added selfless healer to buffs, added stack count to buffs and skills. Did not implement skills for Prot yet.
-- HoJ.ADDON_VERSION = 0.9; -- Warlord's of Draenor. Removed skills that no longer exists (inquisition, blinding light)
-- HoJ.ADDON_VERSION = 0.91; -- Warlord's of Draenor, 6.1 skill priority adjustments.
-- HoJ.ADDON_VERSION = 0.92; -- Warlord's of Draenor, 6.2 priorities from Summonstone
HoJ.MAJOR_VERSION = 0; -- Battle for Azeroth, 8.1 Tides of Vengeance
-- HoJ.MINOR_VERSION = 93; 
-- HoJ.BUILD_VERSION = 1; -- fixed one of the cooldown button timer texts not being set
-- HoJ.BUILD_VERSION = 2; -- fixed a bug where the crusader strike icon was incorrect if the add-on loaded while in prot spec.
-- HoJ.BUILD_VERSION = 3; -- resetting spell icons was resetting empyrean power to not show as divine storm
-- HoJ.BUILD_VERSION = 4; -- removed debug outputs, first beta
-- HoJ.BUILD_VERSION = 5; -- hammer of wrath fixes, added Shield of Vengeance
-- HoJ.BUILD_VERSION = 6; -- selfless healer cooldown
-- HoJ.BUILD_VERSION = 7; -- BUG: sometimes a rotation skill is shown when everything is on cooldown
-- HoJ.BUILD_TYPE = "beta";
HoJ.MINOR_VERSION = 94; 
--HoJ.BUILD_VERSION = 0; -- Very basic Protection rotation
-- HoJ.BUILD_VERSION = 1; -- Very basic Protection mitigation cooldowns
-- HoJ.BUILD_VERSION = 2; -- Amortized update logic to smooth out cost per frame.
HoJ.BUILD_VERSION = 3; -- renamed some frames, added cooldowns
HoJ.BUILD_TYPE = "beta";

HoJ.SPELL_BUFF = 1;			-- buffs a unit
HoJ.SPELL_SHAPESHIFT = 2;	-- buffs self
HoJ.SPELL_SKILL = 3;		-- not a buff, usually part of rotation
HoJ.SPELL_TALENT = 4;		-- not a buff, usually part of rotation, but acquired via talents
HoJ.SPELL_TALENT_BUFF = 5;	-- a buff that is a spell and talented (selfless healer)

HoJ.RANGE_NONE = 0;
HoJ.RANGE_SELF = 1;
HoJ.RANGE_MELEE = 2;
HoJ.RANGE_10_YARDS = 3;
HoJ.RANGE_12_YARDS = 4;
HoJ.RANGE_20_YARDS = 5;
HoJ.RANGE_30_YARDS = 6;
HoJ.RANGE_40_YARDS = 7;

local HOJ_ROLE_HEALING = 1;
local HOJ_ROLE_TANK = 2;
local HOJ_ROLE_DPS = 3;
local HOJ_ROLE_UNKNOWN = 4;

HoJ.REACTION_THRESHOLD = 5; -- 4 will only show skill bar for red mobs, 5 will show skill bar for yellow mobs

HoJ.SKILL_USE_PREDICTION_TIME = 0.1;	-- we can select to use a skill that will be available this many seconds in the future

HoJ.HOJ_ROLE_NAMES = { "HEALING", "TANK", "DPS", "UNKNOWN" };
HoJ.STR_TAB = "	";
HoJ.STR_WHITE = "|cffffffff";
HoJ.STR_YELLOW = "|cffffff00";
HoJ.STR_GREEN = "|cff00ff00";
HoJ.STR_PURPLE = "|cffff00ff";
HoJ.STR_RED = "|cffff0000";
HoJ.STR_ORANGE = "|cffff7f00";
HoJ.STR_LIGHTGRAY = "|cffcfcfcf";
HoJ.STR_GRAY = "|cff7f7f7f";
HoJ.STR_WHISPER = "|cffff7fff";
HoJ.STR_CHANNEL = "|cffffc0c0";

HoJ._needsReset = true;	-- resets occur on role switches

HoJ._enabledForRole = true;

HoJ._role = HOJ_ROLE_UNKNOWN;

HoJ.ADDON_NAME = HoJ.STR_YELLOW .. "Hammer of Justas" .. HoJ.STR_WHITE;
HoJ.MSG_PREFIX = HoJ.STR_YELLOW .. "Hammer of Justas: " .. HoJ.STR_WHITE;
HoJ.UPDATE_INTERVAL = 1.0 / 30.0;
HoJ.RENDER_INTERVAL = 1.0 / 60.0;

-- don't update everything on the same frame
HoJ.UPDATE_START = 0
HoJ.UPDATE_ROTATION = HoJ.UPDATE_START
HoJ.UPDATE_COOLDOWNS = 1
HoJ.UPDATE_BUFFS = 2
HoJ.UPDATE_MAX = HoJ.UPDATE_BUFFS
HoJ._updateType = HoJ.UPDATE_START

HoJ._nextUpdateTime = 0.0
HoJ._nextRenderTime = 0.0

HoJ._isPaladin = false;
HoJ._inCombat = false;
HoJ._level = 0;
HoJ._holyPower = 0;
HoJ.MAX_HOLY_POWER = 5;
HoJ._useHolyPower = false;	-- true when a holy power skill should be used
HoJ._playerName = nil;
HoJ._playerRaceId = 0

HoJ._playerHealth = 1
HoJ._playerMaxHealth = 1
HoJ._playerHealthRatio = 1.0
HoJ._predictedHealth = 1
HoJ._predictedHealthTime = 0.0
HoJ._predictedDeathTime = 0.0
HoJ._predictedNearDeathTime = 0.0

HoJ._damagePerSec = 0
HoJ._healsPerSec = 0
HoJ._damagePerSecRatio = 0.0
HoJ._healsPerSecRatio = 0.0
HoJ._largestRecentDamageRatio = 0.0

HoJ._nextHealthTime = 0.0
HoJ._showHealthTime = 0.0	-- set to a time in the future to see health spam

HoJ.HEALTH_SAMPLE_SECONDS = 3 -- number of seconds we keep health history for
HoJ.HEALTH_SAMPLES_PER_SEC = 30	-- number of times per second that we sample health (make sure it's less than HoJ.UPDATE_INTERVAL)

HoJ._frameNum = 0;
HoJ._updateElapsedTime = 0.0;
HoJ._renderElapsedTime = 0.0;
HoJ._updateBuffsTime = 0.0; -- when to next check buffs
HoJ._updateCooldownsTime = 0.0 -- when to next check cooldowns
--HoJ._divinePleaTriggerPercentage = 0.75;	-- trigger when we have less than this % of max mana
HoJ._blockedNames = { };--["testname"] = "testname" };
--HoJ._blockedNames = { ["Mienyu"] = "Mienyu", ["Justas"] = "Justas" };
HoJ._ignoreHistory = { };
HoJ._silentIgnore = false;
HoJ._stealthIgnore = true;
HoJ._stealthIgnoreMessage = HoJ.STR_CHANNEL .. "[3. WorldDefense] : " .. HoJ.STR_YELLOW .. "Isle of Quel'Danas is under attack!";
HoJ._maxGlowAlpha = 0.5;
HoJ._minGlowAlpha = 0.1;
HoJ._glowAlpha = HoJ._minGlowAlpha;
HoJ._glowScale = 1.025;
HoJ._skillCDAlpha = 0.25;		-- alpha for spell icons on cooldown
HoJ._skillNormalAlpha = 0.75;	-- alpha for skill icons not on cooldown
HoJ._events = { };
HoJ._HoPo = { nil, nil, nil, nil, nil };
HoJ._HoPoGlow = { nil, nil, nil, nil, nil };
HoJ._curTime = GetTime();
HoJ._gcdRemaining = 0.0;
HoJ._gcdDuration = 0.0; -- we cache this because we can't just query it when a spell isn't active
HoJ._minHolyPowerForInquisition = 2;
-- WODONLY
-- HoJ._sacredShieldWarningTime = 2;

HoJ._targetExists = false;
HoJ._targetCanAttack = false;
HoJ._targetCanBeInterrupted = false;
HoJ._targetHealth = 0.0;
HoJ._targetHealthRatio = 0.0;
HoJ._targetLevel = 1;
HoJ._targetClassification = "";
HoJ._targetIsTough = false;

HoJ._targetInRange = { };	-- one entry for each HoJ.RANGE_ type, indicating if the target is within that range

HoJ._showSkillCDs = true;
HoJ._showCooldownsFrame = true;
HoJ._showBuffsFrame = true;
HoJ._showTemplarsVerdictFrame = true;
HoJ._showHolyPowerFrame = true;
HoJ._showInquisitionFrame = true;
HoJ._showRotationFrame = true;

HoJ._enableProfiling = false 
HoJ._profileThreshold = 0.3 -- threshold in ms before showing CPU time

HoJ._enabled = true;

HoJ.INCOMING_DPS_CHECK_INTERVAL = 2.0;
HoJ.AVENGING_WRATH_INCOMING_DPS_RATIO = -0.10;	-- incoming DPS less than this ratio will trigger avenging wrath icon
HoJ.AVENGING_WRATH_HEALTH_THRESHOLD = 0.5;
HoJ.PROT_GUARDIAN_INCOMING_DPS_RATIO = -0.2; -- incoming DPS less than this will trigger guardian of ancient kings icon

HoJ._buffs = { };	-- list of currently active buffs
HoJ._buffFrames = { };

HoJ._cooldowns = { };	-- list of skills to execute in order of priority / cooldown
HoJ._cooldownFrames = { };

HoJ._rotationSkills = { }; -- list of best skills (only ever 1)
HoJ._rotationFrames = { };

HoJ._windowFrames = { };  -- list of all HoJ-specific window frames
HoJ._omniCC = false; -- true if omniCC is enabled (it will show it's own timer)

local HOJ_INQ_EXPIRED		= 0;	-- in combat without inquisition on!
local HOJ_INQ_NEARLY_EXPIRED= 1;	-- in combat but inquisition is expiring
local HOJ_INQ_EXPIRING 		= 2;	-- in combat but inquisition is expiring
local HOJ_INQ_ON 			= 3;	-- in combat and inquisition is on
local HOJ_INQ_NOT_IN_COMBAT	= 4;	-- not in combat, so ignore inquisition
HoJ._inqState = HOJ_INQ_NOT_IN_COMBAT;

local SCROLLBAR_WIDTH = 16;
local SCROLLBAR_HEIGHT = 12;
local EDGE_SIZE = 8;


-- -----------------------------------------------------------------------------
-- HoJ:GetVersionString
-- -----------------------------------------------------------------------------
function HoJ:GetVersionString()
	local version = "" .. HoJ.MAJOR_VERSION .. "." .. HoJ.MINOR_VERSION .. "." .. HoJ.BUILD_VERSION .. " " .. HoJ.BUILD_TYPE;
	return version;
end

-- -----------------------------------------------------------------------------
-- HoJ:DebugMsg
-- -----------------------------------------------------------------------------
function HoJ:DebugMsg( text )
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Warn
-- -----------------------------------------------------------------------------
function HoJ:Warn( text )
	-- should be commented out for releases
	if ( DEFAULT_CHAT_FRAME ) then
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_RED .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Print
-- -----------------------------------------------------------------------------
function HoJ:Print( text ) 
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Msg
-- -----------------------------------------------------------------------------
function HoJ:Msg( text ) 
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.MSG_PREFIX .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellName
-- -----------------------------------------------------------------------------
function HoJ:GetSpellName( spellID )
	name = GetSpellInfo( spellID );
	if ( name == nil ) then
		HoJ:DebugMsg( "Got nil spell name for id " .. spellID );
	end
	return name;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellRank
-- -----------------------------------------------------------------------------
function HoJ:GetSpellRank( spellID )
	_, rank = GetSpellInfo( spellID );
	return rank;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIcon
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIcon( spellID )
	_, _, icon = GetSpellInfo( spellID );
	return icon;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIndex
-- this tries to find the index of a spell in the spellbook, which is not the
-- same as the spell ID of a spell. For some reason the spell index is needed
-- to find if a target is in range of a spell, rather than the spell ID.
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIndex( spellInfo )
	-- tabs are the side tabs which are currently "General", "Retribution", "Holy", and "Protection"
	-- there is no way to know how many spells are on a single tab, but it won't be over 48 for paladins in 8.1
	local numSpellTabs = GetNumSpellTabs();
	--HoJ:DebugMsg( "Found " .. numSpellTabs .. " spellbook tabs" );
	local max = numSpellTabs * 96;
	for i = 1, max do
		local spellName, spellSubName = GetSpellBookItemName( i, BOOKTYPE_SPELL )
		if ( spellName == nil ) then
			--HoJ:DebugMsg( "No spellbook info found for index " .. i );
			return -1;
	   	else
			-- use spellName and spellSubName here
			-- HoJ:DebugMsg( "Spell #" .. i .. ": '" .. spellName .. "'" )
			if ( spellInfo.name == spellName ) then
				--HoJ:DebugMsg( "Found spellbook index " .. i .. " for spell id " .. spellInfo.name );
				return i;
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_InitSpellInfo
-- -----------------------------------------------------------------------------
function HoJ_InitSpellInfo()
	-- HoJ:DebugMsg( "InitSpellInfo" );
	for key, value in pairs( HoJ.SpellInfo ) do
		-- HoJ:DebugMsg( value.name );
		local name, rank, icon = GetSpellInfo( value.id );
		if ( value.iconId ~= nil ) then
			-- HoJ:DebugMsg( "Getting iconId " .. value.iconId .. " for " .. value.name );
			_, _, icon = GetSpellInfo( value.iconId );
		end
		-- For some reason, if we query Crusader Strike while prot, we get a weird texture id
		-- so when we switch roles we go ahead and re-init all of the spell infos.
		-- HoJ:DebugMsg( "  id = " .. value.id );
		-- HoJ:DebugMsg( "  old icon = " .. value.icon .. ", new icon = " .. icon );
		value.name = name;		
		value.rank = rank;
		value.icon = icon;
		value.spellIndex = HoJ:GetSpellIndex( value );
	end		
end

HoJ.SpellInfo = {
--WODONLY
--[[
	["sacredShield"] = {
		-- 20925 is the id for the buff that appears every 6 seconds
		-- 65148 is the id for the buff that appears every 6 seconds
		id = 20925,
		name = HoJ:GetSpellName( 20925 ), -- used to be 53601
		rank = HoJ:GetSpellRank( 20925 ),
		icon = HoJ:GetSpellIcon( 20925 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20925 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,	-- for skills only, if false, don't show
		showCheckbox = true,	-- true to show a checkbox in the options menu - for some stuff that's no longer used it doesn't make sense for this to be true.
		showCharges = false,
		text = nil,
		minLevel = 45,
		cooldownRemaining = 0,
		priority = 0
	},
	["artOfWar"] = {
		id = 59578,
		name = HoJ:GetSpellName( 59578 ),
		rank = HoJ:GetSpellRank( 59578 ),
		icon = HoJ:GetSpellIcon( 59578 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 59578 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 50,
		cooldownRemaining = 0,
		priority = 0
	},
--]]
--[[
	["divinePurpose"] = {
		id = 90174,
		name = HoJ:GetSpellName( 90174 ),
		rank = HoJ:GetSpellRank( 90174 ),
		icon = HoJ:GetSpellIcon( 90174 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 90174 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 50,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
	["divineCrusader"] = {
		id = 144595,
		name = HoJ:GetSpellName( 144595 ),
		rank = HoJ:GetSpellRank( 144595 ),
		icon = HoJ:GetSpellIcon( 144595 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 144595 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 50,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},	
	["supplication"] = {
		id = 94686,
		name = HoJ:GetSpellName( 94686 ),
		rank = HoJ:GetSpellRank( 94686 ),
		icon = HoJ:GetSpellIcon( 94686 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 94686 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 50,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
]]
--[[	
	["divinePlea"] = {
		id = 54428,
		name = HoJ:GetSpellName( 54428 ),
		rank = HoJ:GetSpellRank( 54428 ),
		icon = HoJ:GetSpellIcon( 54428 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 54428 ),
		spellType = HoJ.SPELL_BUFF,
		showOnBar = true,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 46,
		cooldownRemaining = 0,
		priority = 0
	},
--]]	
	["inquisition"] = {
		id = 84963,
		name = HoJ:GetSpellName( 84963 ),
		rank = HoJ:GetSpellRank( 84963 ),
		icon = HoJ:GetSpellIcon( 84963 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		showOnBar = true,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 7,
		talentCol = 3
	},
	["holyAvenger"] = {
		id = 105809,
		name = HoJ:GetSpellName( 105809 ),
		rank = HoJ:GetSpellRank( 105809 ),
		icon = HoJ:GetSpellIcon( 105809 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 75,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},	
	["avengingWrath"] = {
		id = 31884,
		name = HoJ:GetSpellName( 31884 ),
		rank = HoJ:GetSpellRank( 31884 ),
		icon = HoJ:GetSpellIcon( 31884 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = true,
		text = nil,
		minLevel = 80,
		cooldownRemaining = 0,
		priority = 2000,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},	
	["crusade"] = {
		id = 231895,
		name = HoJ:GetSpellName( 231895 ),
		rank = HoJ:GetSpellRank( 231895 ),
		icon = HoJ:GetSpellIcon( 231895 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = true,
		text = nil,
		minLevel = 80,
		cooldownRemaining = 0,
		priority = 2000,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 7,
		talentCol = 2
	},		
	["divineProtection"] = {
		id = 498,
		name = HoJ:GetSpellName( 498 ),
		rank = HoJ:GetSpellRank( 498 ),
		icon = HoJ:GetSpellIcon( 498 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 26,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},	
	["ardentDefender"] = {
		id = 31850,
		name = HoJ:GetSpellName( 31850 ),
		rank = HoJ:GetSpellRank( 31850 ),
		icon = HoJ:GetSpellIcon( 31850 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 70,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},	
	["shieldOfTheRighteous"] = {
		id = 132403,
		name = HoJ:GetSpellName( 132403 ),
		rank = HoJ:GetSpellRank( 132403 ),
		icon = HoJ:GetSpellIcon( 132403 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 20,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},		
	["consecration"] = {
		id = 26573,
		name = HoJ:GetSpellName( 26573 ),
		rank = HoJ:GetSpellRank( 26573 ),
		icon = HoJ:GetSpellIcon( 26573 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT, -- only a talent for ret
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 34,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 1,
		range = HoJ.RANGE_10_YARDS,
		talentRow = 4,	
		talentCol = 2,
		lastCastTime = 0,
		duration = 12
	},	
	["guardianOfAncientKings"] = {
		id = 86659,
		name = HoJ:GetSpellName( 86659 ),
		rank = HoJ:GetSpellRank( 86659 ),
		icon = HoJ:GetSpellIcon( 86659 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 70,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["divineShield"] = {
		id = 642,
		name = HoJ:GetSpellName( 642 ),
		rank = HoJ:GetSpellRank( 642 ),
		icon = HoJ:GetSpellIcon( 642 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 18,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["righteousFury"] = {
		id = 25780,
		name = HoJ:GetSpellName( 25780 ),
		rank = HoJ:GetSpellRank( 25780 ),
		icon = HoJ:GetSpellIcon( 25780 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 25780 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 12,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},	
-- WODONLY
--[[	
	["sealOfTruth"] = {
		id = 31801,
		name = HoJ:GetSpellName( 31801 ),
		rank = HoJ:GetSpellRank( 31801 ),
		icon = HoJ:GetSpellIcon( 31801 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 31801 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 24,
		cooldownRemaining = 0,
		priority = 0
	},
	["sealOfRighteousness"] = {
		id = 20154,
		name = HoJ:GetSpellName( 20154),
		rank = HoJ:GetSpellRank( 20154 ),
		icon = HoJ:GetSpellIcon( 20154 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20154 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 42,
		cooldownRemaining = 0,
		priority = 0
	},	
	["sealOfJustice"] = {
		id = 20164,
		name = HoJ:GetSpellName( 20164),
		rank = HoJ:GetSpellRank( 20164 ),
		icon = HoJ:GetSpellIcon( 20164 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20164 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 70,
		cooldownRemaining = 0,
		priority = 0
	},
	["sealOfInsight"] = {
		id = 20165,
		name = HoJ:GetSpellName( 20165 ),
		rank = HoJ:GetSpellRank( 20165 ),
		icon = HoJ:GetSpellIcon( 20165 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20165 ),
		spellType = HoJ.SPELL_SHAPESHIFT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 32,
		cooldownRemaining = 0,
		priority = 0
	},
	["blessingOfMight"] = {
		id = 19740,
		name = HoJ:GetSpellName( 19740 ),
		rank = HoJ:GetSpellRank( 19740 ),
		icon = HoJ:GetSpellIcon( 19740 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 19740 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 81,
		cooldownRemaining = 0,
		priority = 0
	},
--]]
--WODONLY	
--[[	
	["blessingOfKings"] = {
		id = 20217,
		name = HoJ:GetSpellName( 20217 ),
		rank = HoJ:GetSpellRank( 20217 ),
		icon = HoJ:GetSpellIcon( 20217 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0
	},
]]	
	["blessingOfKings"] = { -- actually called "greater blessing of kings"
		id = 203538,
		name = HoJ:GetSpellName( 203538 ),
		rank = HoJ:GetSpellRank( 203538 ),
		icon = HoJ:GetSpellIcon( 203538 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},
	["blessingOfWisdom"] = { -- actually called "greater blessing of wisdom"
		id = 203539,
		name = HoJ:GetSpellName( 203539 ),
		rank = HoJ:GetSpellRank( 203539 ),
		icon = HoJ:GetSpellIcon( 203539 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},
	["blessingOfProtection"] = {
		id = 1022,
		name = HoJ:GetSpellName( 1022 ),
		rank = HoJ:GetSpellRank( 1022 ),
		icon = HoJ:GetSpellIcon( 1022 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},		
	["hammerOfWrath"] = {
		id = 24275,
		name = HoJ:GetSpellName( 24275 ),
		rank = HoJ:GetSpellRank( 24275 ),
		icon = HoJ:GetSpellIcon( 24275 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 24275 ),
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 38,
		cooldownRemaining = 0,
		priority = 800,
		hoPoGen = 1,
		range = HoJ.RANGE_30_YARDS,
		talentRow = 2,
		talentCol = 3
	},
	["rebuke"] = {
		id = 96231,
		name = HoJ:GetSpellName( 96231 ),
		rank = HoJ:GetSpellRank( 96231 ),
		icon = HoJ:GetSpellIcon( 96231 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 96231 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 1,
		hoPoGen = 0,
		range = HoJ.RANGE_MELEE
	},
	["crusaderStrike"] = {
		id = 35395,
		name = HoJ:GetSpellName( 35395 ),
		rank = HoJ:GetSpellRank( 35395 ),
		icon = HoJ:GetSpellIcon( 35395 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 1,
		cooldownRemaining = 0,
		priority = 600,
		hoPoGen = 1,
		range = HoJ.RANGE_MELEE
	},
-- WODONLY	
--[[	
	["exorcism"] = {
		id = 879,
		name = HoJ:GetSpellName( 879),
		rank = HoJ:GetSpellRank( 879 ),
		icon = HoJ:GetSpellIcon( 879 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 879 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 46,
		cooldownRemaining = 0,
		priority = 700
	},
--]]	
	["judgment"] = {
		id = 20271,
		name = HoJ:GetSpellName( 20271 ),
		rank = HoJ:GetSpellRank( 20271 ),
		icon = HoJ:GetSpellIcon( 20271 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 20271 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 5,
		cooldownRemaining = 0,
		priority = 400,
		hoPoGen = 1,
		range = HoJ.RANGE_30_YARDS
	},
	["templarsVerdict"] = {
		id = 85256,
		name = HoJ:GetSpellName( 85256 ),
		rank = HoJ:GetSpellRank( 85256 ),
		icon = HoJ:GetSpellIcon( 85256 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 85256 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 1,
		cooldownRemaining = 0,
		priority = 1400,
		hoPoGen = -3,
		range = HoJ.RANGE_MELEE
	},	
	["finalVerdict"] = {
		id = 157048,
		name = HoJ:GetSpellName( 157048 ),
		rank = HoJ:GetSpellRank( 157048 ),
		icon = HoJ:GetSpellIcon( 157048 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 157048 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1400,
		hoPoGen = -3,
		range = HoJ.RANGE_MELEE
	},
	["finalVerdictBuff"] = {	-- this skill is also the buff...
		id = 157048,
		name = HoJ:GetSpellName( 157048 ),
		rank = HoJ:GetSpellRank( 157048 ),
		icon = HoJ:GetSpellIcon( 157048 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 157048 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 100,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},	
	["divineStorm"] = {	
		id = 53385,
		name = HoJ:GetSpellName( 53385 ),
		rank = HoJ:GetSpellRank( 53385 ),
		icon = HoJ:GetSpellIcon( 53385 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 53385 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1300,
		hoPoGen = -3,
		range = HoJ.RANGE_12_YARDS
	},
	["executionSentence"] = {	
		id = 163250,
		name = HoJ:GetSpellName( 163250 ),
		rank = HoJ:GetSpellRank( 163250 ),
		icon = HoJ:GetSpellIcon( 163250 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1500,
		hoPoGen = -3,
		range = HoJ.RANGE_20_YARDS,
		talentRow = 1,
		talentCol = 3
	},
	["lightsHammer"] = {	
		id = 114158,
		name = HoJ:GetSpellName( 114158 ),
		rank = HoJ:GetSpellRank( 114158 ),
		icon = HoJ:GetSpellIcon( 114158 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 114158 ),
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1500,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},	
	["globalCooldown"] = {	
		id = 61304, -- must be something on the GCD -- Crusader Strike for now
		name = HoJ:GetSpellName( 61304 ),
		rank = HoJ:GetSpellRank( 61304 ),
		icon = HoJ:GetSpellIcon( 61304 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 1,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
	["selflessHealer"] = {
		id = 85804,
		name = HoJ:GetSpellName( 85804 ),
		rank = HoJ:GetSpellRank( 85804 ),
		icon = HoJ:GetSpellIcon( 85804 ),
		spellToCast = 19750, -- cast as flash of light
		spellIndex = -1,  -- HoJ:GetSpellIndex( 85804 ),
		spellType = HoJ.SPELL_TALENT_BUFF,
		isProcd = true,
		showOnBar = true, 	-- set to false so we don't show on the buffs bar by default, because we show on cooldowns due to the special circumstances
		showCheckbox = true,
		showCharges = false,
		showAsBuff = false,
		text = nil,
		minLevel = 0,
		cooldownRemaining = 0,
		priority = 500,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE,
		talentRow = 6,
		talentCol = 1,
		maxBuffCount = 4
	},
	["layOnHands"] = {
		id = 633,
		name = HoJ:GetSpellName( 633 ),
		rank = HoJ:GetSpellRank( 633 ),
		icon = HoJ:GetSpellIcon( 633 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 16,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},
	["lightOfTheProtector"] = {
		id = 184092,
		name = HoJ:GetSpellName( 184092 ),
		rank = HoJ:GetSpellRank( 184092 ),
		icon = HoJ:GetSpellIcon( 184092 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 75,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},	
	["handOfTheProtector"] = {
		id = 213652,
		name = HoJ:GetSpellName( 213652 ),
		rank = HoJ:GetSpellRank( 213652 ),
		icon = HoJ:GetSpellIcon( 213652 ),
		spellIndex = -1,
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 75,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS,
		talentRow = 5,
		talentCol = 3
	},		
	["hammerOfJustice"] = {
		id = 853,
		name = HoJ:GetSpellName( 853 ),
		rank = HoJ:GetSpellRank( 853 ),
		icon = HoJ:GetSpellIcon( 853 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 7,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_10_YARDS
	},
	["fistOfJustice"] = {
		id = 234299,
		name = HoJ:GetSpellName( 234299 ),
		rank = HoJ:GetSpellRank( 234299 ),
		icon = HoJ:GetSpellIcon( 234299 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 45,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_10_YARDS,
		talentRow = 3,
		talentCol = 1
	},
	["blindingLight"] = {
		id = 115750,
		name = HoJ:GetSpellName( 115750 ),
		rank = HoJ:GetSpellRank( 115750 ),
		icon = HoJ:GetSpellIcon( 115750 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 45,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_10_YARDS,
		talentRow = 3,
		talentCol = 3
	},	
	["flashOfLight"] = {
		id = 19750,
		name = HoJ:GetSpellName( 19750 ),
		rank = HoJ:GetSpellRank( 19750 ),
		icon = HoJ:GetSpellIcon( 19750 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS
	},
	["giftOfTheNaaru"] = {
		id = 59542,
		name = HoJ:GetSpellName( 59542 ),
		rank = HoJ:GetSpellRank( 59542 ),
		icon = HoJ:GetSpellIcon( 59542 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_40_YARDS,
		requiredRaceId = 11
	},
	["wakeOfAshes"] = {
		id = 255937,
		name = HoJ:GetSpellName( 255937 ),
		rank = HoJ:GetSpellRank( 255937 ),
		icon = HoJ:GetSpellIcon( 255937 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 5,
		range = HoJ.RANGE_12_YARDS,
		talentRow = 4,
		talentCol = 3
	},
	["bladeOfJustice"] = {
		id = 184575,
		name = HoJ:GetSpellName( 184575 ),
		rank = HoJ:GetSpellRank( 184575 ),
		icon = HoJ:GetSpellIcon( 184575 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 20,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 2,
		range = HoJ.RANGE_12_YARDS
	},
	["empyreanPower"] = {
		id = 286390,
		name = HoJ:GetSpellName( 286390 ),
		rank = HoJ:GetSpellRank( 286390 ),
		iconId = 53385, -- shows the divine storm icon in the skill window instead of empyrean power
		icon = HoJ:GetSpellIcon( 53385 ),	
		spellIndex = -1, 
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 120,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_12_YARDS
	},	
	["wordOfGlory"] = {
		id = 210191,
		name = HoJ:GetSpellName( 210191 ),
		rank = HoJ:GetSpellRank( 210191 ),
		icon = HoJ:GetSpellIcon( 210191 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 5,
		range = HoJ.RANGE_40_YARDS,
		talentRow = 6,
		talentCol = 3
	},
	["shieldOfVengeance"] = {
		id = 184662,
		name = HoJ:GetSpellName( 184662 ),
		rank = HoJ:GetSpellRank( 184662 ),
		icon = HoJ:GetSpellIcon( 184662 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 10,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_30_YARDS
	},	
	["avengersShield"] = {
		id = 31935,
		name = HoJ:GetSpellName( 31935 ),
		rank = HoJ:GetSpellRank( 31935 ),
		icon = HoJ:GetSpellIcon( 31935 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 10,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["hammerOfTheRighteous"] = {
		id = 53595,
		name = HoJ:GetSpellName( 53595 ),
		rank = HoJ:GetSpellRank( 53595 ),
		icon = HoJ:GetSpellIcon( 53595 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_SKILL,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 10,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF
	},
	["blessedHammer"] = {
		id = 204019,
		name = HoJ:GetSpellName( 204019 ),
		rank = HoJ:GetSpellRank( 204019 ),
		icon = HoJ:GetSpellIcon( 204019 ),
		spellIndex = -1, 
		spellType = HoJ.SPELL_TALENT,
		isProcd = false,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 15,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_SELF,
		talentRow = 1,
		talentCol = 3
	},
	
-- WODONLY
--[[	
	["crusadersFury"] = {
		id = 165442,
		name = HoJ:GetSpellName( 165442 ),
		rank = HoJ:GetSpellRank( 165442 ),
		icon = HoJ:GetSpellIcon( 165442 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 165442 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = true,
		showCheckbox = true,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 1000
	},	
--]]	
	["blazingContempt"] = {
		id = 166831,
		name = HoJ:GetSpellName( 166831 ),
		rank = HoJ:GetSpellRank( 166831 ),
		icon = HoJ:GetSpellIcon( 166831 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 166831 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = true,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 100,
		cooldownRemaining = 0,
		priority = 900,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
--[[
	["spiritBeastBlessing"] = { -- hunter buff provides same benefit as might + 5% crit
		id = 128997,
		name = HoJ:GetSpellName( 128997 ),
		rank = HoJ:GetSpellRank( 128997 ),
		icon = HoJ:GetSpellIcon( 128997 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 128997 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},	
	["graceOfAir"] = { -- shaman buff provides same benefit as might + 5% haste
		id = 116956,
		name = HoJ:GetSpellName( 116956 ),
		rank = HoJ:GetSpellRank( 116956 ),
		icon = HoJ:GetSpellIcon( 116956 ),
		spellIndex = -1,  -- HoJ:GetSpellIndex( 116956 ),
		spellType = HoJ.SPELL_BUFF,
		isProcd = false,
		showOnBar = false,
		showCheckbox = false,
		showCharges = false,
		text = nil,
		minLevel = 30,
		cooldownRemaining = 0,
		priority = 0,
		hoPoGen = 0,
		range = HoJ.RANGE_NONE
	},
]]	
};

HoJ._combatAuraSpellInfo = HoJ.SpellInfo.retributionAura; -- spell info for the aura we want to check for in combat

local sliderBackdrop  = {
	bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
	edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
	tile = true, tileSize = 8, edgeSize = 8,
	insets = { left = 3, right = 3, top = 6, bottom = 6 }
};

local listBackdrop = {
  -- path to the background texture
  bgFile = "",
  -- path to the border texture
  edgeFile = "Interface/Buttons/UI-SliderBar-Border",
  -- true to repeat the background texture to fill the frame, false to scale it
  tile = true,
  -- size (width or heighet) of the square repeating background tiles (in pixels)
  tileSize = 8,
  -- thickness of edge segments and square size of edge corners (in pixels)
  edgeSize = EDGE_SIZE,
  -- distance from the edges of the frame to those of the background texture (in pixels)
  insets = {
	left = 3,
	right = 3,
	top = 3,
	bottom = 3
  }
};

local SAVED_VARIABLES = "HoJSavedVars";

local DEFAULTS = {
	["blockedNames"] = { },
	["silentIgnore"] = false,
	["stealthIgnore"] = false,
	["showSkillCDs"] = true,
	["showSkillsBar"] = true,
	["showBuffsBar"] = true,
	["spellInfo"] = { }
};

HoJ.savedVars = { };

-- -----------------------------------------------------------------------------
-- HoJ:InitSavedVariables
-- -----------------------------------------------------------------------------
function HoJ:InitSavedVariables()
	HoJ.savedVars = DEFAULTS;

	if ( _G[SAVED_VARIABLES] == nil ) then
		HoJ:DebugMsg( "Using defaults." );
	else
		HoJ:DebugMsg( "Using saved variables." );

		HoJ.savedVars = _G[SAVED_VARIABLES];
		for key, value in pairs( HoJ.SpellInfo ) do
			local savedSpellInfo = HoJ.savedVars["spellInfo"];
			if ( savedSpellInfo[key] == nil ) then
				HoJ:DebugMsg( "Adding saved var " .. key .. ", " .. value.name );
				savedSpellInfo[key] = { showOnBar = true };
			-- else
			--	HoJ:DebugMsg( "Found saved var " .. key .. " = " .. tostring( savedSpellInfo[key].showOnBar ) );
			end
		end
	end

	HoJ:ReadSavedVariables();

	_G[SAVED_VARIABLES] = HoJ.savedVars;
end

-- -----------------------------------------------------------------------------
-- HoJ:SetDefaults
-- -----------------------------------------------------------------------------
function HoJ:SetDefaults()
	HoJ.savedVars = DEFAULTS;
	_G[SAVED_VARIABLES] = HoJ.savedVars;
	HoJ:ReadSavedVariables();
end

-- -----------------------------------------------------------------------------
-- HoJ:ReadBool
-- -----------------------------------------------------------------------------
function HoJ:ReadBool( name, default )
	if ( name == nil ) then
		return;
	end
	
	if ( HoJ.savedVars[name] ~= nil ) then
		-- HoJ:DebugMsg( "Read " .. name .. " = " .. tostring( HoJ.savedVars[name] ) );
		return HoJ.savedVars[name];
	else
		-- HoJ:DebugMsg( "Read " .. name .. " = nil, set to default of " .. tostring( default ) );
		return default;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ReadSavedVariables
-- -----------------------------------------------------------------------------
function HoJ:ReadSavedVariables()
	--HoJ:DebugMsg( "ReadSavedVariables" );
	if ( HoJ.savedVars.spellInfo ~= nil ) then
		for key, value in pairs( HoJ.savedVars.spellInfo ) do
			--HoJ:DebugMsg( "Reading var " .. key );
			if ( HoJ.SpellInfo[ key ] ~= nil ) then -- don't error if there's an old var
				HoJ.SpellInfo[ key ].showOnBar = value.showOnBar;
			end
		end
	end
	
	-- ignore list
	HoJ:ReadIgnoreVariables();
	
	-- misc
	HoJ._showSkillCDs = HoJ:ReadBool( "showSkillCDs", true );
	HoJ._showCooldownsFrame = HoJ:ReadBool( "showCooldownsFrame", true );
	HoJ._showBuffsFrame = HoJ:ReadBool( "showBuffsFrame", true );	
	HoJ._showRotationFrame = HoJ:ReadBool( "showRotationFrame", true );	
	HoJ._showTemplarsVerdictFrame = HoJ:ReadBool( "showTemplarsVerdictFrame", true );		
	HoJ._showHolyPowerFrame = HoJ:ReadBool( "showHolyPowerFrame", true );	
	HoJ._showInquisitionFrame = HoJ:ReadBool( "showInquisitionFrame", true );	
end

-- -----------------------------------------------------------------------------
-- HoJ:WriteBool
-- -----------------------------------------------------------------------------
function HoJ:WriteBool( name, var, default )
	if ( var == nil ) then
		HoJ:DebugMsg( "Boolean var " .. name .. " is nil!" );
		HoJ.savedVars[name] = default;
	else
		--HoJ:DebugMsg( "Wrote " .. name .. " = " .. tostring( var ) );
		HoJ.savedVars[name] = var;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:WriteSavedVariables
-- -----------------------------------------------------------------------------
function HoJ:WriteSavedVariables()
	--HoJ:DebugMsg( "WriteSavedVariables" );
	for key, value in pairs( HoJ.SpellInfo ) do
		--HoJ:DebugMsg( "Writing var " .. key );
		local t = { showOnBar = value.showOnBar; }
		local v = HoJ.savedVars["spellInfo"];
		if ( v == nil ) then
			HoJ:DebugMsg( "v = nil" );
		end
		v[ key ] = t;
	end

	-- ignore list
	HoJ:WriteIgnoreVariables();
	
	-- misc
	HoJ:WriteBool( "showSkillCDs", HoJ._showSkillCDs, true );
	HoJ:WriteBool( "showSkillsFrame", HoJ.showCooldownsFrame, true );
	HoJ:WriteBool( "showBuffsFrame", HoJ._showBuffsFrame, true );
	HoJ:WriteBool( "showRotationFrame", HoJ._showRotationFrame, true );
	HoJ:WriteBool( "showTemplarsVerdictFrame", HoJ._showTemplarsVerdictFrame, true );
	HoJ:WriteBool( "showHolyPowerFrame", HoJ._showHolyPowerFrame, true );
	HoJ:WriteBool( "showInquisitionFrame", HoJ._showInquisitionFrame, true );
end

-----------------------------------------------------------------------------
-- HoJ:ReadIgnoreVariables
-- -----------------------------------------------------------------------------
function HoJ:ReadIgnoreVariables()
	if ( HoJ.savedVars["blockedNames"] ~= nil ) then
		HoJ._blockedNames = HoJ.savedVars["blockedNames"];
	end
	if ( HoJ.savedVars["silentIgnore"] ~= nil ) then
		HoJ._silentIgnore = HoJ.savedVars["silentIgnore"];
	end
	if ( HoJ.savedVars["stealthIgnore"] ~= nil ) then
		HoJ._stealthIgnore = HoJ.savedVars["stealthIgnore"];
	end
end

-----------------------------------------------------------------------------
-- HoJ:WriteIgnoreVariables
-- -----------------------------------------------------------------------------
function HoJ:WriteIgnoreVariables()
	HoJ.savedVars["blockedNames"] = HoJ._blockedNames;
	HoJ.savedVars["silentIgnore"] = HoJ._silentIgnore;
	HoJ.savedVars["stealthIgnore"] = HoJ._stealthIgnore;
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateRole
-- -----------------------------------------------------------------------------
function HoJ:UpdateRole()
	local specId = GetSpecialization();
	
	-- HoJ:DebugMsg( "UpdateRole(): specId = " .. specId );
	if ( specId == 3 ) then
		HoJ._role = HOJ_ROLE_DPS;
	elseif ( specId == 2 ) then
		HoJ._role = HOJ_ROLE_TANK;
	elseif ( specId == 1 ) then
		HoJ._role = HOJ_ROLE_HEALING;
	else
		HoJ._role = HOJ_ROLE_UNKNOWN;
	end
end

-- -----------------------------------------------------------------------------
-- Slash command handling
-- -----------------------------------------------------------------------------
SLASH_HAMMEROFJUSTAS1, SLASH_HAMMEROFJUSTAS2 = '/hoj', '/hammerofjustas';
function SlashCmdList.HAMMEROFJUSTAS( msg, editbox )
	local command, rest = msg:match( "^(%S*)%s*(.-)$" );
	local name = rest:match( "^(%S*)%s*(%a-)" );

	-- HoJ:Print( "command = '" .. command .. "', rest = '" .. rest .. "'." );
	-- HoJ:Print( "name = '" .. name .. "'." );	
	
	if ( command == "defaults" ) then
		HoJ:SetDefaults();
		HoJ:Msg( "Reset to defaults." );
	elseif ( command == "ignore" or command == "add" ) then
		HoJ._blockedNames[ name ] = name;
		HoJ.WriteBlockedNames();
		HoJ:Msg( "Added " .. name .. " to whisper ignore list." );
	elseif ( command == "unignore" or command == "remove" ) then
		HoJ._blockedNames[ name ] = nil;
		HoJ.WriteBlockedNames();
		HoJ:Msg( "Removed " .. name .. " from whisper ignore list." );
	elseif ( command == "list" ) then
		local count = 1;
		HoJ:Msg( "Ignore List:" );
		for key, value in pairs( HoJ._blockedNames ) do
			HoJ:Print( HoJ.STR_WHITE .. "  " .. count .. ": " .. value );
			count = count + 1;
		end
	elseif ( command == "silent" ) then
		if ( HoJ._silentIgnore == true ) then
			HoJ._silentIgnore = false;
			HoJ:Msg( "Silent Ignore is OFF." );
		else
			HoJ._silentIgnore = true;
			HoJ:Msg( "Silent Ignore is ON." );
		end
	elseif ( command == "stealth" ) then		
		if ( HoJ._stealthIgnore == true ) then
			HoJ._stealthIgnore = false;
			HoJ:Msg( "Stealth Ignore is OFF." );
		else
			HoJ._stealthIgnore = true;
			HoJ:Msg( "Stealth Ignore is ON." );
		end	
	elseif ( command == "history" ) then
		HoJ:Msg( "Ignore History:" );
		for key, value in pairs( HoJ._ignoreHistory ) do
			HoJ:Print( HoJ.STR_WHITE .. "From " .. key .. ": " );
			for i, historyMsg in ipairs( HoJ._ignoreHistory[ key ] ) do
				HoJ:Print( HoJ.STR_WHISPER .. "  " .. historyMsg );
			end
		end
	elseif ( command == "clear" ) then
		HoJ._ignoreHistory = { };
		HoJ:Print( "History cleared." );
	elseif ( command == "prof" or command == "profile" ) then
		if ( HoJ._enableProfiling ) then
			HoJ._enableProfiling = false;
			HoJ:Msg( "Profiling turned OFF." );
		else
			HoJ._enableProfiling = true;
			HoJ:Msg( "Profiling turned ON." );
		end
elseif ( command == "thresh" or command == "threshold" ) then
		HoJ._profileThreshold = tonumber( name )
		HoJ:Msg( "Profiling threshold set to " .. HoJ._profileThreshold );
	elseif ( command == "disable" ) then
		HoJ._enabled = false;
		HoJ._needsReset = true;
		HoJ:Msg( "Update disabled." );
	elseif ( command == "enable" ) then
		HoJ._enabled = true;
		HoJ:Msg( "Update enabled." );		
	else 
		HoJ:Msg( "slash commands:" );
		HoJ:Print( HoJ.STR_WHITE .. "  ignore <name> - ignores whispers from the specified player." );
		HoJ:Print( HoJ.STR_WHITE .. "  add <name> - same as ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  unignore <name> - stops ignoring whispers from the specified player." );
		HoJ:Print( HoJ.STR_WHITE .. "  remove <name> - same as unignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  list - lists all currently ignored players." );
		HoJ:Print( HoJ.STR_WHITE .. "  silent - toggle silent ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  stealth - toggle stealth ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  profile - toggle addon profiling." );
		HoJ:Print( HoJ.STR_WHITE .. "  disable - disable the addon's update." );
		HoJ:Print( HoJ.STR_WHITE .. "  enable - enable the addon's update." );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:EnterCombat
-- -----------------------------------------------------------------------------
function HoJ:EnterCombat( ... )
	--HoJ:DebugMsg( "HoJ:EnterCombat" );
	HoJ._inCombat = true;

	-- intiate the inquisition cooldown because we may have turned on inquisition	
	-- outside of combat.
--	HoJ:StartInquisitionCooldown();
end

-- -----------------------------------------------------------------------------
-- HoJ:LeaveCombat
-- -----------------------------------------------------------------------------
function HoJ:LeaveCombat( ... )
	--HoJ:DebugMsg( "HoJ:LeaveCombat" );
	HoJ._inCombat = false;
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIcon
-- -----------------------------------------------------------------------------
function HoJ:AddIcon( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" );
	text = text or "";
	
	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:AddIcon - Icon list is NIL" );
	end
	
	if ( not spellInfo.showOnBar ) then
		HoJ:RemoveIcon( list, spellInfo );
		return;
	end
	
	-- HoJ:DebugMsg( "AddIcon( " .. spellInfo.name .. " )" );
	local si = spellInfo;
	si.text = text;
	table.insert( list, si );
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconExclusive
-- -----------------------------------------------------------------------------
function HoJ:AddIconExclusive( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" );
	text = text or "";
	
	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:AddIcon - Icon list is NIL" );
	end

	for i = 1, #list do
		if ( list[i].name == spellInfo.name ) then
			return
		end
	end
	
	if ( not spellInfo.showOnBar ) then
		HoJ:RemoveIcon( list, spellInfo );
		return;
	end
	
	-- HoJ:DebugMsg( "AddIcon( " .. spellInfo.name .. " )" );
	local si = spellInfo;
	si.text = text;
	table.insert( list, si );
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconDebug
-- just adds debug text where the cooldown text would be
function HoJ:AddIconDebug( list, spellInfo, debugText )
	HoJ:AddIcon( list, spellInfo, "" )
end

-- -----------------------------------------------------------------------------
-- HoJ:RemoveIcon
-- -----------------------------------------------------------------------------
function HoJ:RemoveIcon( list, spellInfo )
	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:RemoveIcon - Icon list is NIL" );
	end
	-- HoJ:DebugMsg( "RemoveIcon( " .. spellInfo.name .. " )" );
	for i, v in pairs( list ) do
		if ( v == spellInfo ) then
			table.remove( list, i );
			break;
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ClearIconList
-- -----------------------------------------------------------------------------
function HoJ:ClearIconList( list )
	for i, v in pairs( list ) do
		list[ i ] = nil;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SetIconFrame
-- -----------------------------------------------------------------------------
function HoJ:SetIconFrame( v, spellInfo )
	if ( v == nill ) then
		return; -- there are more skills than frames
	end
	
	v.frame:EnableMouse( true );
	--v.frame:Enable();
	v.frame:Show();
	-- HoJ:DebugMsg( "Enable mouse for " .. v.frame:GetName() );

	v.spellInfo = spellInfo;
	if ( v.texture ~= nil ) then
		v.texture:SetTexture( spellInfo.icon );
		v.texture:Show();
	end

	if ( v.text ~= nil ) then	
		if ( spellInfo.text == nil or spellInfo.text == "" ) then
			v.text:Hide();
			if ( v.texture ~= nil ) then
				v.texture:SetAlpha( HoJ._skillNormalAlpha );
			end
		else
			local fontFile = GameFontHighlightLarge:GetFont();
			v.text:SetFont( fontFile, 12, "THICKOUTLINE" );
			v.text:SetText( spellInfo.text );
			v.text:Show();
			if ( spellInfo.spellType == HoJ.SPELL_SKILL or spellInfo.spellType == HoJ.SPELL_TALENT ) then
				if ( v.texture ~= nil ) then
					v.texture:SetAlpha( HoJ._skillCDAlpha );
				end
			end
		end
	end

	-- if the frame has a cooldown, set it
	local cooldownFrame = v.cooldown
	if ( cooldownFrame ~= nil ) then
		local start, duration, enabled = GetSpellCooldown( spellInfo.id )
		if ( start > 0 and duration > 0 ) then
			cooldownFrame:Show()
			cooldownFrame:SetCooldown( start, duration )
		end
	end	


	-- we cannot use SetAttribute() on a secure frame while in combat
	if ( not HoJ._inCombat ) then
		if ( spellInfo.spellType == HoJ.SPELL_BUFF or 
			 spellInfo.spellType == HoJ.SPELL_TALENT_BUFF or
			 spellInfo.spellType == HoJ.SPELL_SHAPESHIFT or
			 spellInfo.spellType == HoJ.SPELL_SKILL or
			 spellInfo.spellType == HoJ.SPELL_TALENT ) then
			-- HoJ:DebugMsg( "spell " .. spellInfo.name );
			v.frame:SetAttribute( "type", "spell" );
			v.frame:SetAttribute( "unit", "player" );
			if ( v.spellToCast ~= nil ) then
				-- some buffs are meant to affect another spell (i.e. Selfless Healer modifies cast time of Flash of Light)
				-- though, due to combat restrictions, this isn't actually castable if it is added to the bar during combat (which it usually is).
				v.frame:SetAttribute( "spell", HoJ:GetSpellName( spellInfo.spellToCast ) );
			else
				v.frame:SetAttribute( "spell", spellInfo.name );
			end
		else
			HoJ:DebugMsg( "UNKNOWN " .. spellInfo.name );
			--v.frame:SetAttribute( "type", "" );
			--v.frame:SetAttribute( "unit", "" );
			--v.frame:SetAttribute( "spell", "" );
		end	
	end
end

function SkillCompare( tbl, a, b )
	local cdra = tbl[a].cooldownRemaining;
	local cdrb = tbl[b].cooldownRemaining;
	
	if ( cdra <= 0.00001 ) then
		if ( cdrb <= 0.00001 ) then
			return ( tbl[a].priority > tbl[b].priority );
		else
			return true;
		end
	elseif ( cdrb <= 0.00001 ) then
		-- tbl[a] must be on cooldown or we'd have gone into the first if
		return false;
	else
		-- both are on cooldown, sort by cooldown remaining time
		return ( cdra < cdrb );
	end	
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateIcons
-- -----------------------------------------------------------------------------
function HoJ:UpdateIcons( list, frameList, sortList ) 
	-- clear all icon frames
	for key, value in pairs( frameList ) do
		if ( value ~= nil ) then
			if ( value.texture ~= nil ) then
				value.texture:SetTexture( nil );
				value.texture:Hide();
			end
			value.spellInfo = nil;
			if ( value.text ~= nil ) then
				value.text:Hide();
			end
			-- calling EnableMouse( false ) causes the OnMouseUp / OnClick (for mouse up)
			-- not to fire at all, so instead we call EnableMouse( false ) only for
			-- buttons that have no spell assigned.
			--value.frame:EnableMouse( false );
		end
	end	
	
	-- create an array for sorting the skills
	sortedKeys = { };
	for key in pairs( list ) do
		table.insert( sortedKeys, key );
	end
	if ( sortList ) then
		table.sort( sortedKeys, function( a, b ) return SkillCompare( list, a, b ) end );
	end
	
	-- go through the active icons array and set up a frame for each
	local frameNum = 1;
	for i, n in ipairs( sortedKeys ) do
		--HoJ:DebugMsg( "n = " .. n );
		--HoJ:DebugMsg( "spellInfo name = " .. spellInfo.name );
		HoJ:SetIconFrame( frameList[frameNum], list[n] );
		frameNum = frameNum + 1;
	end

	for i, value in ipairs( frameList ) do
		if ( i >= frameNum ) then
			value.frame:EnableMouse( false );
		end
	end

	return frameNum > 1;
end

function HoJ:UpdateFrameIcons( frame, list, subFrameList, sortList )
	if ( HoJ:UpdateIcons( list, subFrameList, sortList ) ) then
		--frame:EnableMouse( true );
		-- HoJ:DebugMsg( "EnableMouse for " .. frame:GetName() );
	else
		--frame:EnableMouse( false );
		-- HoJ:DebugMsg( "DisableMouse for " .. frame:GetName() );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitBuffForName
-- -----------------------------------------------------------------------------
function HoJ:UnitBuffForName( unit, buffName )
	local i = 1;
	local name, icon, count = UnitAura( unit, i );
	while name do
		if ( name == buffName ) then
			return name, icon, count;
		end
		i = i + 1;
		name, icon, count = UnitBuff( unit, i );
	end
	return name, icon, count;
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBuff
-- -----------------------------------------------------------------------------
function HoJ:FindBuff( buffName )
	local i = 1;
	local name = UnitBuff( "player", i );
	while name do 
		-- HoJ:DebugMsg( "buffName " .. i .. " " .. name );
		if ( name == buffName ) then
			return true;
		end
		i = i + 1;
		name = UnitBuff( "player", i );
	end
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBuffIndex
-- -----------------------------------------------------------------------------
function HoJ:FindBuffIndex( buffName )
	local i = 1;
	local name = UnitBuff( "player", i );
	while name do 
		-- HoJ:DebugMsg( "buffName " .. i .. " " .. name );
		if ( name == buffName ) then
			return i;
		end
		i = i + 1;
		name = UnitBuff( "player", i );
	end
	return -1;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingWithPrediction
-- gets the cooldown remaining, including GCD if GCD is in progress
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingWithPrediction( spellId, predictAhead ) 
	local start, duration, enabled = GetSpellCooldown( spellId );

	-- we don't really care about enabled... that just tells us if the spell / buff is active
	if ( start > 0 and duration > 0 ) then
		-- we "look into the future" to just past the time when the current GCD is done,
		-- to see what will be usable then
		local remaining = ( start + duration ) - ( HoJ._curTime + predictAhead );
		if ( remaining < 0.0 ) then
			remaining = 0.0;
		end
		return remaining;
	end
	return 0.0;	
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemaining
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemaining( spellId ) 
	return HoJ:GetCDRemainingWithPrediction( spellId, 0.0 );
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingAfterGCD
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingAfterGCD( spellId, extraPrediction ) 
	-- we "look into the future" to just past the time when the current GCD is done,
	-- to see what will be usable then	
	return HoJ:GetCDRemainingWithPrediction( spellId, HoJ._gcdRemaining + extraPrediction );	
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffTimeRemainingWithGCD
-- returns timeStarted, buffDuration, timeRemaining
-- -----------------------------------------------------------------------------
function HoJ:GetBuffTimeRemainingWithGCD( buffName )
	local buffIndex = HoJ:FindBuffIndex( buffName );
	local timeRemaining = 0;
	local start = 0;
	local duration = 0;
	local expires = 0;
	if ( buffIndex >= 0 ) then
		_, _, _, _, duration, expires = UnitBuff( "player", buffIndex );
		if ( expires ) then
			--HoJ:DebugMsg( "time = " .. HoJ._curTime .. ", gcd = " .. HoJ._gcdRemaining );
			timeRemaining = expires - HoJ._curTime - HoJ._gcdRemaining;
			start = expires - duration;
		end
	end
	return start, duration, timeRemaining;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffTimeRemainingWithoutGCD
-- -----------------------------------------------------------------------------
function HoJ:GetBuffTimeRemainingWithoutGCD( buffName )
	local buffIndex = HoJ:FindBuffIndex( buffName );
	local timeRemaining = 0;
	local start = 0;
	local duration = 0;
	local expires = 0;
	if ( buffIndex >= 0 ) then
		_, _, _, _, duration, expires = UnitBuff( "player", buffIndex );
		if ( expires ) then
			--HoJ:DebugMsg( "time = " .. HoJ._curTime .. ", gcd = " .. HoJ._gcdRemaining );
			timeRemaining = expires - HoJ._curTime;
			start = expires - duration;
		end
	end
	return start, duration, timeRemaining;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetTimeRemainingForLongestDurationBuff
--
-- sacred shield has two buffs, the 30 second one and the 6 second recurring one.
-- we can't actually tell them apart by id, so
-- -----------------------------------------------------------------------------
function HoJ:GetTimeRemainingForLongestDurationBuff( buffName )
	local i = 1;
	local longestTime = -1;
	local expireTime = 0;
	local name, _, _, _, duration, expires = UnitBuff( "player", i );
	while name do 
		if ( duration > longestTime and name == buffName ) then
			-- HoJ:DebugMsg( "buff " .. name .. " has duration " .. duration .. " and expires at time " .. expires .. " seconds." );
			longestTime = duration;
			expireTime = expires;
		end
		i = i + 1;
		name, _, _, _, duration, expires = UnitBuff( "player", i );
	end
	if ( longestTime >= 0 ) then
		local timeRemaining = expireTime - HoJ._curTime - HoJ._gcdRemaining;
		return timeRemaining;
	end
	return 0.0;
end

-- -----------------------------------------------------------------------------
-- HoJ:InRange
-- -----------------------------------------------------------------------------
function HoJ:InRange( spellInfo, unit )
	--local inRange = IsSpellInRange( spellInfo.name, unit );
	--local inRange = IsSpellInRange( spellInfo.spellIndex, "spellbook", unit );
	local inRange = false;
	if ( spellInfo.spellType == HoJ.SPELL_TALENT ) then
		inRange = IsSpellInRange( spellInfo.spellIndex, "spell", unit );
	else
		inRange = IsSpellInRange( spellInfo.name, unit );
	end

	if ( inRange == nil ) then
		--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = nil" );
		return false;
	elseif ( inRange == 0 ) then
		--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = 0" );
		return false;
	end	
	--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = true" );
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:BoolToStr
-- -----------------------------------------------------------------------------
function HoJ:BoolToStr( boolVar )
	if ( boolVar ) then
		return "true";
	end
	return "false";
end

-- -----------------------------------------------------------------------------
-- HoJ:WillOverflowHoPo
-- -----------------------------------------------------------------------------
function HoJ:WillOverflowHoPo( spellInfo, curHoPo, maxHoPo )
	local overflow = ( curHoPo + spellInfo.hoPoGen ) - maxHoPo;
	-- HoJ:DebugMsg( spellInfo.name .. ": overflow = " .. overflow );
	return overflow;
end	

-- -----------------------------------------------------------------------------
-- HoJ:IsOffCooldown
-- -----------------------------------------------------------------------------
function HoJ:IsOffCooldown( spellInfo, cdThreshold )
	local cdRemaining = HoJ:GetCDRemainingAfterGCD( spellInfo.id, cdThreshold );
	-- HoJ:DebugMsg( "spell " .. spellInfo.name .. ": cdRemaining = " .. cdRemaining .. ", GCD Remaining = " .. HoJ._gcdRemaining );
	if ( cdRemaining == nil ) then
		HoJ:DebugMsg( "cdRemaining == nil!" )
		return false;
	elseif ( cdThreshold == nil ) then
		HoJ:DebugMsg( "cdThreshold == nil!" )
		return false;
	end
	if ( cdRemaining <= cdThreshold ) then
		return true;
	end
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsSkillUsable
-- -----------------------------------------------------------------------------
function HoJ:IsSkillUsable( spellInfo, cdThreshold )
	if ( not HoJ._targetInRange[spellInfo.range] ) then
		-- HoJ:DebugMsg( spellInfo.name .. " is not in range" )
		return false;
	end
	if ( not HoJ:IsOffCooldown( spellInfo, cdThreshold ) ) then
		return false;
	end
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsSkillUsableWithoutOverflow
-- -----------------------------------------------------------------------------
function HoJ:IsSkillUsableWithoutOverflow( spellInfo, cdThreshold, curHoPo, maxHoPo )
	if ( not HoJ:IsSkillUsable( spellInfo, cdThreshold ) ) then
		return false;
	end
	if ( HoJ:WillOverflowHoPo( spellInfo, curHoPo, maxHoPo ) > 0 ) then
		return false;
	end
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateProcdBuff
-- -----------------------------------------------------------------------------
function HoJ:UpdateProcdBuff( name, icon, count, duration, expires )
	if ( name == nil ) then
		HoJ:DebugMsg( "UpdateProcdBuff called with nil name!" );
		return
	end

	for key, value in pairs( HoJ.SpellInfo ) do
		if ( value.isProcd and value.showOnBar and 
			 ( value.showAsBuff == nil or value.showAsBuff == true ) and
			 ( value.spellType == HoJ.SPELL_BUFF or ( value.spellType == HoJ.SPELL_TALENT_BUFF and HoJ:IsTalentSelected( value ) ) ) and 
			 value.name ~= nil and value.name == name ) then

			--HoJ:DebugMsg( "UpdateProcdBuff: " .. name );

			timeRemaining = expires - HoJ._curTime - HoJ._gcdRemaining;
			--start = expires - duration;				
			if ( count > 1 ) then	
				if ( expires ) then
					HoJ:AddIcon( HoJ._buffs, value, string.format("%i/%i", count, timeRemaining ) );
				else
					HoJ:AddIcon( HoJ._buffs, value, string.format("%i", count ) );
				end
			elseif ( expires ) then					
				if ( timeRemaining <= 3.0 ) then
					HoJ:AddIcon( HoJ._buffs, value, string.format( "%.1f", timeRemaining ) );
				else 
					HoJ:AddIcon( HoJ._buffs, value, string.format( "%i", timeRemaining ) );
				end
			else
				HoJ:AddIcon( HoJ._buffs, value, "" );
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckProcdBuffs
-- -----------------------------------------------------------------------------
function HoJ:CheckProcdBuffs()
	if ( not HoJ._showBuffsFrame ) then
		return;
	end

	local index = 1;	
	repeat
		local name, icon, count, debuffType, duration, expires = UnitBuff( "player", index );
		if ( name == nil ) then
			break;
		end
		HoJ:UpdateProcdBuff( name, icon, count, duration, expires );
		index = index + 1;
	until ( index > 40 )
end

-- -----------------------------------------------------------------------------
-- HoJ:ShowSkillsList
-- -----------------------------------------------------------------------------
function HoJ:ShowSkillsList( prefix, list )
	local skillsOrder = "";
	for i, v in pairs( list ) do
		if ( i ~= 1 ) then
			skillsOrder = skillsOrder .. ", ";
		end
		if ( v.text == nil ) then
			skillsOder = skillsOrder .. "nil";
		else
			skillsOrder = skillsOrder .. v.text;
		end
	end
	HoJ:DebugMsg( prefix .. skillsOrder );
end

-- -----------------------------------------------------------------------------
-- HoJ:IsTalentSelected
-- -----------------------------------------------------------------------------
function HoJ:IsTalentSelected( spellInfo )
	if ( spellInfo.talentRow == nil or spellInfo.talentCol == nil ) then
		if ( spellInfo.spellType ~= HoJ.SPELL_TALENT ) then
			HoJ:DebugMsg( "IsTalentSelected() called on non-talent " .. spellInfo.name );
		else
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " missing row and column" );
		end
		return true;
	end

	local _, _, _, selected, available = GetTalentInfo( spellInfo.talentRow, spellInfo.talentCol, 1 );
	return ( available and selected );
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckInquisition
-- -----------------------------------------------------------------------------
function HoJ:CheckInquisition()
	if ( HoJ._level < HoJ.SpellInfo.inquisition.minLevel or not HoJ._inCombat or not HoJ:IsTalentSelected( HoJ.SpellInfo.inquisition ) ) then
		HoJ._inqState = HOJ_INQ_NOT_IN_COMBAT;
		return 0.0, 0.0, 0.0
	end
	local inqStart, inqDuration, inqTimeRemaining = HoJ:GetBuffTimeRemainingWithoutGCD( HoJ.SpellInfo.inquisition.name );

	local cdText = string.format( "%.1f", inqTimeRemaining );
	HoJ_InqTimer:SetText( cdText );

	if ( inqTimeRemaining <= 0 ) then
		--HoJ:DebugMsg( "HOJ_INQ_EXPIRED" );
		HoJ._inqState = HOJ_INQ_EXPIRED;
		HoJ_InqTimer:SetTextColor( 1.0, 0.0, 0.0 );
		HoJ_InqTimer:SetText( "0.0" );
		HoJ:AddIcon( HoJ._cooldowns, HoJ.SpellInfo.inquisition, "0" );
	elseif ( inqTimeRemaining <= 5 ) then
		--HoJ:DebugMsg( "HOJ_INQ_NEARLY_EXPIRED" );
		HoJ._inqState = HOJ_INQ_NEARLY_EXPIRED;
		HoJ_InqTimer:SetTextColor( 1.0, 0.0, 0.0 );
	elseif ( inqTimeRemaining <= 10 ) then
		--HoJ:DebugMsg( "HOJ_INQ_EXPIRING" );
		HoJ._inqState = HOJ_INQ_EXPIRING;
		HoJ_InqTimer:SetTextColor( 1.0, 1.0, 0.0 );
	else
		--HoJ:DebugMsg( "HOJ_INQ_ON" );
		HoJ_InqTimer:SetTextColor( 1.0, 1.0, 1.0 );
		HoJ._inqState = HOJ_INQ_ON;
	end
	
	return inqStart, inqDuration, inqTimeRemaining;
end

-- -----------------------------------------------------------------------------
-- HoJ:StartInquisitionCooldown
-- -----------------------------------------------------------------------------
function HoJ:StartInquisitionCooldown()
	if ( HoJ._inqState == HOJ_INQ_NOT_IN_COMBAT ) then
		return;
	end

	local inqStart, inqDuration, inqTimeRemaining = HoJ:CheckInquisition();
	--HoJ:DebugMsg( "inqStart = " .. inqStart .. ", inqDuration = " .. inqDuration );
	CooldownFrame_SetTimer( HoJ_InqFrameCooldown, inqStart, inqDuration, 1.0 );	
	
	-- find the OmniCC timer frame and hide it so it doesn't overlay our timer
	if ( HoJ._omniCC ) then
		local frame = EnumerateFrames();
		while frame do
			if ( frame:GetParent() == HoJ_InqFrame ) then
				if ( frame:GetName() == nil ) then							
					--HoJ:DebugMsg( "A unnamed frame is parented to HoJ_InqFrame." );
					frame:SetAlpha( 0.0 );	-- go away, omniCC
					break
				end
			end
			frame = EnumerateFrames( frame );
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateInquisitionFrames
-- -----------------------------------------------------------------------------
function HoJ:UpdateInquisitionFrames( self )
	--HoJ:DebugMsg( "UpdateInquisitionFrames" );
	if ( HoJ._showInquisitionFrame and HoJ._role == HOJ_ROLE_DPS and HoJ._enabledForRole ) then
		if ( HoJ._inqState == HOJ_INQ_NOT_IN_COMBAT ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_NOT_IN_COMBAT" );
			HoJ_InqFrameCooldown:SetAlpha( 0.0 );
			HoJ_InqTexture:SetAlpha( 0.0 );
			HoJ_InqTimer:SetAlpha( 0.0 );
		elseif ( HoJ._inqState == HOJ_INQ_EXPIRED ) then
			-- animate inquisition texture alpha
			local inqAlpha = finalAlpha + 0.25;
			if ( inqAlpha > 1.0 ) then
				inqAlpha = 1.0;
			end
			--HoJ:DebugMsg( "Render HOJ_INQ_EXPIRED " .. string.format( "%.2f", inqAlpha ) );			
			HoJ_InqFrameCooldown:SetAlpha( inqAlpha );
			HoJ_InqTexture:SetAlpha( inqAlpha );
			HoJ_InqTimer:SetAlpha( inqAlpha );
		elseif ( HoJ._inqState == HOJ_INQ_ON ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_ON" );
			HoJ_InqFrameCooldown:SetAlpha( 0.1 );
			HoJ_InqTexture:SetAlpha( 0.1 );
			HoJ_InqTimer:SetAlpha( 0.2 );
		elseif ( HoJ._inqState == HOJ_INQ_NEARLY_EXPIRED ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_*" );
			HoJ_InqFrameCooldown:SetAlpha( 1.0 );
			HoJ_InqTexture:SetAlpha( 1.0 );
			HoJ_InqTimer:SetAlpha( 1.0 );			
		elseif ( HoJ._inqState == HOJ_INQ_EXPIRING ) then
			--HoJ:DebugMsg( "Render HOJ_INQ_*" );
			HoJ_InqFrameCooldown:SetAlpha( 0.4 );
			HoJ_InqTexture:SetAlpha( 0.4 );
			HoJ_InqTimer:SetAlpha( 0.5 );
		end
	else
		HoJ_InqFrameCooldown:SetAlpha( 0.0 );
		HoJ_InqTexture:SetAlpha( 0.0 );
		HoJ_InqTimer:SetAlpha( 0.0 );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetHealthRatio
-- -----------------------------------------------------------------------------
function HoJ:GetHealthRatio( unitName )
	local maxHealth = UnitHealthMax( unitName );
	local curHealth = UnitHealth( unitName );
	return curHealth / maxHealth;
end

-- -----------------------------------------------------------------------------
-- HoJ:FindActiveMitigationSkills
-- table of flags for each mitigation skill, indexed by spell id
--	{
--		isActive,	-- whether or not the spell is active
--	}
-- -----------------------------------------------------------------------------
function HoJ:FindActiveMitigationSkills( activeSkills )
	-- set all active skills active flag to false
	for key, skill in pairs( activeSkills ) do
		skill.isActive = false
	end

	local i = 1;
	local name, _, _, _, _, _, _, _, _, spellId = UnitAura( "player", i )
	while name do
		-- HoJ:DebugMsg( "Skill " .. name .. " is active" )
		if ( spellId ~= nil ) then
			-- HoJ:DebugMsg( "Has spellId = " .. spellId )
			local activeSkill = activeSkills[spellId]
			if ( activeSkill ~= nil ) then
				-- HoJ:DebugMsg( "Found " .. name .. " in activeSkills" )
				activeSkill.isActive = true
			end
		end
		i = i + 1;
		name, _, _, _, _, _, _, _, _, spellId = UnitAura( "player", i )
	end
end

--[[
function HoJ:FindActiveMitigationSkills()
	local i = 1;
	local name = UnitBuff( "player", i );
	while name do 
		local spellInfo = self.SpellInfo[name]
		if ( spellInfo ~= nil ) then
			spellInfo.isActive = true

		for j, mitigation do
			if ( name == mitigation.spellInfo.name ) then
				mitigation.isActive = true
				self:DebugMsg( "Mitigation " .. name .. " is active." )
			else
				mitigation.isActive = false
			end
		end
		i = i + 1;
		name = UnitBuff( "player", i );
	end	
end
]]

-- -----------------------------------------------------------------------------
-- HoJ:SelectMitigationSkill
-- expects an array of table items of format:
--	{
--		spellInfo,	-- info about the spell
--		replacedByTalent,	-- option, can be nil, otherwise specifies the spellInfo of the talent that replaces this spell 
--      physicalOnly-- optional, can be nil
--	}
-- NOTE: the list MUST be in priority order because the loop short-circuits once a
-- mitigation is found, so the order IS the priority
-- -----------------------------------------------------------------------------
function HoJ:SelectMitigationSkill( mitigationSkills, listName )
	if ( listName ~= nil ) then
		self:DebugMsg( HoJ.STR_GREEN .. "Checking mitigations: " .. listName )
	end

	for i, mitigation in ipairs( mitigationSkills ) do
		-- self:DebugMsg( "Checking mitigation skill " .. mitigation.spellInfo.name )
		if ( mitigation.spellInfo.requiredRaceId == nil or mitigation.spellInfo.requiredRaceId == self._playerRaceId ) then
			if ( not self._activeMitigationList[mitigation.spellInfo.id].isActive ) then
				if ( mitigation.replacedByTalent == nil or not self:IsTalentSelected( mitigation.replacedByTalent ) ) then
					local isSelected = true
					if ( mitigation.spellInfo.type == HoJ.SPELL_TALENT or mitigation.spellInfo.type == HoJ.SPELL_TALENT ) then
						 isSelected = self:IsTalentSelected( mitigation.spellInfo )
					 	-- self:DebugMsg( "Talent " .. mitigation.spellInfo.name .. " selected = " .. self:BoolToStr( isSelected ) )
					end
					if ( ( mitigation.physicalOnly ~= nil and HoJ._isTakingPhysicalDamage ) or mitigation.physicalOnly == nil ) then
						if ( isSelected ) then
							if ( self:IsOffCooldown( mitigation.spellInfo, 0.0 ) ) then
								-- self:DebugMsg( HoJ.STR_GREEN .. "Selected mitigation: " .. mitigation.spellInfo.name )
								return mitigation
							end
						end
					end
				end
			end
		end
	end
end

-- list of all mitigation skills and whether or not they are active
HoJ._activeMitigationList = {
	[HoJ.SpellInfo.handOfTheProtector.id] = { isActive = false },
	[HoJ.SpellInfo.lightOfTheProtector.id] = { isActive = false },
	[HoJ.SpellInfo.shieldOfTheRighteous.id] = { isActive = false },
	[HoJ.SpellInfo.giftOfTheNaaru.id] = { isActive = false },
	[HoJ.SpellInfo.ardentDefender.id] = { isActive = false },
	[HoJ.SpellInfo.guardianOfAncientKings.id] = { isActive = false },
	[HoJ.SpellInfo.layOnHands.id] = { isActive = false },
	[HoJ.SpellInfo.blessingOfProtection.id] = { isActive = false },
	[HoJ.SpellInfo.divineShield.id] = { isActive = false },
}

HoJ._lowHealthMitigationList = { }
HoJ._lowHealthMitigationList[1] = { spellInfo = HoJ.SpellInfo.handOfTheProtector }
HoJ._lowHealthMitigationList[2] = { spellInfo = HoJ.SpellInfo.lightOfTheProtector, replacedByTalent = HoJ.SpellInfo.handOfTheProtector }

-- MODERATE MITIGATION SKILL PRIORITY
HoJ._moderateMitigationList = { }
HoJ._moderateMitigationList[1] = { spellInfo = HoJ.SpellInfo.shieldOfTheRighteous }
HoJ._moderateMitigationList[2] = { spellInfo = HoJ.SpellInfo.giftOfTheNaaru }
HoJ._moderateMitigationList[3] = { spellInfo = HoJ.SpellInfo.handOfTheProtector }
HoJ._moderateMitigationList[4] = { spellInfo = HoJ.SpellInfo.lightOfTheProtector, replacedByTalent = HoJ.SpellInfo.handOfTheProtector }
HoJ._moderateMitigationList[5] = { spellInfo = HoJ.SpellInfo.ardentDefender }
HoJ._moderateMitigationList[6] = { spellInfo = HoJ.SpellInfo.guardianOfAncientKings }

-- EMERGENCY MITIGATION SKILL PRIORITY
-- in situations where we're about to die very soon, prefer to
--
-- 1) heal using Hand / Light of the Protector
-- 2) heal using Lay on Hands
-- 3) cast ardent defender (because this will prevent a death and heal to 20%)
-- 4) cast guardian of ancient kings
-- 4) blessing of protection (assuming incoming damage is physical)
-- 5) divine shield (last resort, should be followed by a taunt-and-cancel if tanking)
-- 
-- not really sure about these priorities... in some situations where I'm pretty sure
-- I'm not going to die I prefer to to use Ardent Defender every two minutes as a 
-- damage-reducer only.
HoJ._emergencyMitigationList = { }
HoJ._emergencyMitigationList[1] = { spellInfo = HoJ.SpellInfo.layOnHands }
HoJ._emergencyMitigationList[2] = { spellInfo = HoJ.SpellInfo.handOfTheProtector }
HoJ._emergencyMitigationList[3] = { spellInfo = HoJ.SpellInfo.lightOfTheProtector, replacedByTalent = HoJ.SpellInfo.handOfTheProtector }
HoJ._emergencyMitigationList[4] = { spellInfo = HoJ.SpellInfo.ardentDefender }
HoJ._emergencyMitigationList[5] = { spellInfo = HoJ.SpellInfo.guardianOfAncientKings }
HoJ._emergencyMitigationList[6] = { spellInfo = HoJ.SpellInfo.blessingOfProtection }
HoJ._emergencyMitigationList[7] = { spellInfo = HoJ.SpellInfo.divineShield }

-- -----------------------------------------------------------------------------
-- HoJ:CheckMitigationSkills
-- -----------------------------------------------------------------------------

function HoJ:CheckMitigationSkills()
	if ( not self._showCooldownsFrame ) then
		return;
	end

--	if ( not self._inCombat ) then
--		return;
--	end

	self:FindActiveMitigationSkills( HoJ._activeMitigationList )

	-- if < 50% health, show Light of the Protector
	if ( self._playerHealthRatio < 0.5 ) then
		local lowHealthMitigation = self:SelectMitigationSkill( self._lowHealthMitigationList )
		if ( lowHealthMitigation ) then
			self:AddIcon( self._cooldowns, lowHealthMitigation.spellInfo )
		end
	end

	if ( self._activeMitigationList[self.SpellInfo.divineShield.id].isActive ) then
		-- self:DebugMsg( "Divine Shield is on" );
		return
	end	

	-- moderate damage mitigation
	if ( self._damagePerSecRatio > 0.1 or self._damagePerSecRatio * 5.0 > self._playerHealthRatio ) then
		local moderateMitigation = self:SelectMitigationSkill( self._moderateMitigationList )
		if ( moderateMitigation ) then
			self:AddIconExclusive( self._cooldowns, moderateMitigation.spellInfo )
		end
	end

	-- emergency mitigation
	if ( self._damagePerSecRatio * 2.0 > self._playerHealthRatio ) then
		local emergencyMitigation = self:SelectMitigationSkill( self._emergencyMitigationList )
		if ( emergencyMitigation ) then
			self:AddIconExclusive( self._cooldowns, emergencyMitigation.spellInfo )
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckHolyAvenger
-- -----------------------------------------------------------------------------
function HoJ:CheckShieldOfVengeance( skillsList )
	if ( not self._showBuffsFrame ) then
		return
	end

	if ( not self:IsOffCooldown( self.SpellInfo.shieldOfVengeance, 0.1 ) ) then
		return
	end

	-- don't bother showing Shield of Vengeance if we're not taking much damage
	if ( self._damagePerSecRatio < 0.1 ) then
		return
	end
	local shieldOfVengeanceOn = self:FindBuff( self.SpellInfo.shieldOfVengeance );
	if ( not self._inCombat ) then
		if ( not shieldOfVengeanceOn and self._targetCanAttack ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.shieldOfVengeance );
		end
	else
		if ( not isOn ) then
			self:AddIcon( self._cooldowns, self.SpellInfo.shieldOfVengeance );
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckAvengingWrath
-- -----------------------------------------------------------------------------
function HoJ:CheckAvengingWrath()
	if ( not HoJ._showCooldownsFrame ) then
		return;
	end

	if ( HoJ._level < HoJ.SpellInfo.avengingWrath.minLevel or not HoJ._inCombat ) then
		return;
	end

	local spell = HoJ.SpellInfo.avengingWrath;
	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.crusade ) ) then
		spell = HoJ.SpellInfo.crusade;
	end

	local awIsOn = HoJ:FindBuff( spell.name );
	if ( awIsOn ) then
		-- don't bother showing an icon if Avenging Wrath is currently active
		return;
	end
		
	local cooldownRemaining = HoJ:GetCDRemaining( spell.id );
	local charges, maxCharges = GetSpellCharges( spell.id );

	if ( charges == nil or maxCharges == 1 ) then
		-- without tier 18 2-piece armor bonus
		if ( cooldownRemaining < 2.5 ) then
			HoJ:AddIcon( HoJ._cooldowns, spell );
		end
	else
		-- happens with tier 18 2-piece armor bonus
		if ( charges > 0 ) then
			HoJ:AddIcon( HoJ._cooldowns, spell, string.format( "%i", charges ) );
		elseif ( cooldownRemaining < 2.5 ) then
			HoJ:AddIcon( HoJ._cooldowns, spell, string.format( "%.1f", awCD ) );
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckConsecration
-- -----------------------------------------------------------------------------
--[[
function HoJ:CheckConsecration()
	if ( not HoJ._showBuffsFrame ) then
		return;
	end
	if ( not HoJ._inCombat ) then
		-- HoJ:RemoveIcon( HoJ._buffs, HoJ.SpellInfo.consecration );
		return;
	end
	
	local isOn = HoJ:FindBuff( HoJ.SpellInfo.consecration.name );
	local cdRemaining = HoJ:GetCDRemaining( HoJ.SpellInfo.consecration.id );
	if ( not isOn and cdRemaining <= 0 ) then
		HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.consecration );
	else
		HoJ:RemoveIcon( HoJ._buffs, HoJ.SpellInfo.consecration );
	end
end
--]]

-- -----------------------------------------------------------------------------
-- HoJ:CheckSeals
-- -----------------------------------------------------------------------------
-- WODOONLY - no seals anymore
--[[
function HoJ:CheckSeals()
	if ( not HoJ._showBuffsFrame ) then
		return;
	end
	
	local truthIndex = 1;
	local righteousnessIndex = 2;
	local insightIndex = 3;
	local justiceIndex = 3;
	local justiceActive = false;
	local justiceIsCastable = false;	
	if ( HoJ._role == HOJ_ROLE_DPS ) then
		insightIndex = 4;
		_, _, justiceActive, justiceIsCastable = GetShapeshiftFormInfo( 3 );
	end
	
	local _, _, truthActive, _ = GetShapeshiftFormInfo( truthIndex );
	local _, _, righteousnessActive, _ = GetShapeshiftFormInfo( righteousnessIndex );
	local _, _, insightActive, _ = GetShapeshiftFormInfo( insightIndex );
		
--	if ( truthActive ) then
--		HoJ:DebugMsg( "Seal of Truth is active!" );
--	end
--	if ( righteousnessActive ) then
--		HoJ:DebugMsg( "Seal of Righteousness is active!" );
--	end
--	if ( justiceActive ) then
--		HoJ:DebugMsg( "Seal of Justice is active!" );
--	end
--	if ( insightActive ) then
--		HoJ:DebugMsg( "Seal of Insight is active!" );
--	end
	
	local preferredSeal = HoJ.SpellInfo.sealOfTruth;
	if ( HoJ._role == HOJ_ROLE_HEALING ) then
		preferredSeal = HoJ.SpellInfo.sealOfInsight;
	end
	
	if ( justiceIsCastable ) then
		if ( not truthActive and 
			 not righteousnessActive and
			 not insightActive and
			 not justiceActive ) then
			 HoJ:AddIcon( HoJ._buffs, preferredSeal );
		end
	else
		if ( not truthActive and 
			 not righteousnessActive and
			 not insightActive ) then
			 HoJ:AddIcon( HoJ._buffs, preferredSeal );
		end
	end
end
--]]

-- -----------------------------------------------------------------------------
-- HoJ:CheckRighteousFury
-- -----------------------------------------------------------------------------
-- BFAFIX
--[[
function HoJ:CheckRighteousFury()
	if ( not HoJ._showBuffsFrame ) then
		return;
	end
	local isOn = HoJ:FindBuff( HoJ.SpellInfo.righteousFury.name );
	if ( not isOn ) then
		--HoJ:DebugMsg( "Righteous Fury is not on!" );
		HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.righteousFury );
	else
		--HoJ:DebugMsg( "Righteous Fury is on!" );
		HoJ:RemoveIcon( HoJ._buffs, HoJ.SpellInfo.righteousFury );
	end
end
--]]

-- -----------------------------------------------------------------------------
-- HoJ:CheckBlessings
-- -----------------------------------------------------------------------------
function HoJ:CheckBlessings()
	if ( not HoJ._showBuffsFrame ) then
		return;
	end
	--HoJ:DebugMsg( "CheckBlessings" );
--[[	
	local mightIsOn = false;HoJ:FindBuff( HoJ.SpellInfo.blessingOfMight.name );
	local kingsIsOn = HoJ:FindBuff( HoJ.SpellInfo.blessingOfKings.name );
	local graceIsOn = HoJ:FindBuff( HoJ.SpellInfo.graceOfAir.name );
	local spiritBeastIsOn = HoJ:FindBuff( HoJ.SpellInfo.spiritBeastBlessing.name );
	if ( not mightIsOn and not kingsIsOn and not spiritBeastIsOn ) then
		if ( HoJ._level >= HoJ.SpellInfo.blessingOfMight.minLevel and not mightIsOn ) then
			-- HoJ:DebugMsg( "Adding Blessing of Might icon" );
			HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfMight );
		end		
		if ( HoJ._level >= HoJ.SpellInfo.blessingOfKings.minLevel and not kingsIsOn ) then
			-- HoJ:DebugMsg( "Adding Blessing of Kings icon" );
			HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfKings );
		end
	end
--]]
-- TODO: don't show blessings in a raid or party unless we can check if anyone in the party has our buff
	local kingsIsOn = HoJ:FindBuff( HoJ.SpellInfo.blessingOfKings.name );
	if ( HoJ._level >= HoJ.SpellInfo.blessingOfKings.minLevel and not kingsIsOn ) then
		-- HoJ:DebugMsg( "Adding Blessing of Kings icon" );
		HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfKings );
	end
	local wisdomIsOn = HoJ:FindBuff( HoJ.SpellInfo.blessingOfWisdom.name );
	if ( HoJ._level >= HoJ.SpellInfo.blessingOfWisdom.minLevel and not wisdomIsOn ) then
		-- HoJ:DebugMsg( "Adding Blessing of Wisdom icon" );
		HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfWisdom );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckSelflessHealer
-- -----------------------------------------------------------------------------
function HoJ:CheckSelflessHealer()
	if ( not HoJ:IsTalentSelected( HoJ.SpellInfo.selflessHealer ) ) then
		return;
	end

	local name = nil;
	local icon = nil;
	local count = 0;

	name, icon, count = HoJ:UnitBuffForName( "player", HoJ.SpellInfo.selflessHealer.name );
	if ( name == nil ) then
		--HoJ:DebugMsg( "unitbuff name is nil" );
		return;
	end
	--HoJ:DebugMsg( "unitbuff name is " .. name .. ", count is " .. count );
	if ( count >= HoJ.SpellInfo.selflessHealer.maxBuffCount or HoJ._playerHealthRatio <= 0.5 ) then
		HoJ:AddIcon( HoJ._cooldowns, HoJ.SpellInfo.selflessHealer, string.format("%i", count ) );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes
-- -----------------------------------------------------------------------------
function HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes( curHoPo, maxHoPo )
	--HoJ:DebugMsg( "GetNextHoPoGeneratorWithoutWakeOfAshes" );

	if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.bladeOfJustice, HoJ.SKILL_USE_PREDICTION_TIME, curHoPo, maxHoPo ) ) then
		return HoJ.SpellInfo.bladeOfJustice;
	end

	if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.judgment, HoJ.SKILL_USE_PREDICTION_TIME, curHoPo, maxHoPo ) ) then
		return HoJ.SpellInfo.judgment;
	end

	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.hammerOfWrath ) ) then
		if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.hammerOfWrath, HoJ.SKILL_USE_PREDICTION_TIME, curHoPo, maxHoPo ) ) then
			if ( HoJ._targetHealthRatio <= 0.20 or HoJ:FindBuff( HoJ.SpellInfo.crusade.name ) or HoJ:FindBuff( HoJ.SpellInfo.avengingWrath.name ) ) then
				return HoJ.SpellInfo.hammerOfWrath;
			end
		end
	end

	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.consecration ) ) then
		if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.consecration, HoJ.SKILL_USE_PREDICTION_TIME, curHoPo, maxHoPo ) ) then
			return HoJ.SpellInfo.consecration;
		end
	end

	-- do not use crusader strike if we have >= 3 holy power and the charges are < 2
	currentCharges = GetSpellCharges( HoJ.SpellInfo.crusaderStrike.id );
	if ( HoJ._holyPower < 3 or currentCharges > 1 ) then 
		if ( HoJ:IsSkillUsableWithoutOverflow( HoJ.SpellInfo.crusaderStrike, HoJ.SKILL_USE_PREDICTION_TIME, curHoPo, maxHoPo ) ) then
			return HoJ.SpellInfo.crusaderStrike;
		end
	end

	--HoJ:DebugMsg( "No HoPo gen available" );
	return nil;	
end

-- -----------------------------------------------------------------------------
-- HoJ:GetNextHoPoGenerator
-- -----------------------------------------------------------------------------
function HoJ:GetNextHoPoGenerator( curHoPo )
	--HoJ:DebugMsg( "GetNextHoPoGenerator" );

	local wakeOfAshesSelected = HoJ:IsTalentSelected( HoJ.SpellInfo.wakeOfAshes );
	local wakeOfAshesUsable = HoJ:IsSkillUsable( HoJ.SpellInfo.wakeOfAshes, HoJ.SKILL_USE_PREDICTION_TIME );
	local amountOfWakeOfAshesOverflow = HoJ:WillOverflowHoPo( HoJ.SpellInfo.wakeOfAshes, curHoPo, HoJ.MAX_HOLY_POWER );
	-- HoJ:DebugMsg( "Wake sel = " .. HoJ:BoolToStr( wakeOfAshesSelected ) .. ", use = " .. HoJ:BoolToStr( wakeOfAshesUsable ) .. ", over = " .. amountOfWakeOfAshesOverflow );
	if ( wakeOfAshesSelected ) then
		if ( wakeOfAshesUsable and amountOfWakeOfAshesOverflow <= 0 ) then
			return HoJ.SpellInfo.wakeOfAshes;
		end
	end

	-- if Wake of Ashes is usable but will overflow, then, if possible, choose a skill that will
	-- take us to exactly 3 HoPo so we can use it, then use Wake of Ashes without overflow next.
	local nextSkill = nil;
	if ( wakeOfAshesSelected and wakeOfAshesUsable ) then
		-- if we have exactly 3 HoPo then don't choose a skill here, which will result in the
		-- next chosen skill being Templar's Verdict or Divine Storm, allowing us to use Wake
		-- next without any HoPo lost
		if ( curHoPo == 3 ) then
			return nil;
		end
		-- try to get a skill that won't overflow 3 HoPo
		nextSkill = HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes( curHoPo, 3 );
		if ( nextSkill ~= nil ) then
			--HoJ:DebugMsg( "Gen: " .. nextSkill.name .. " (without Wake)." );
			return nextSkill;
		end
	end

	nextSkill = HoJ:GetNextHoPoGeneratorWithoutWakeOfAshes( curHoPo, HoJ.MAX_HOLY_POWER );
--	if ( nextSkill ~= nil ) then
--		HoJ:DebugMsg( "Gen: " .. nextSkill.name );
--	end
	return nextSkill;
end

-- -----------------------------------------------------------------------------
-- HoJ:ChooseHolyPowerSkill
-- -----------------------------------------------------------------------------
function HoJ:ChooseHolyPowerSkill( skillsList )
	if ( HoJ._holyPower + HoJ.SpellInfo.executionSentence.hoPoGen >= 0 ) then
		if ( HoJ:IsTalentSelected( HoJ.SpellInfo.executionSentence ) ) then
			return HoJ.SpellInfo.executionSentence;
		end
	end

	-- TODO: figure out a way to tell if we should use Divine Storm instead
	if ( HoJ._holyPower + HoJ.SpellInfo.templarsVerdict.hoPoGen >= 0 ) then
		return HoJ.SpellInfo.templarsVerdict;
	end

	return nil;
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckDPSRotation_Ret_8dot1_Internal

-- DPS rotation for Retribution
--[[
Based on Icy-Veins rotation as of 12/27/2018:

The single target rotation consists of alternating Holy Power generation and Holy Power expenditure.

We begin with the priority of Holy Power generation, which you should utilize until you are either at 5 Holy Power, or until the next generator you use would put you over 5 Holy Power (thus wasting the excess).

1. Cast Wake of Ashes.
2. Cast Blade of Justice.
3. Cast Judgment.
4. Cast Hammer of Wrath.
5. Cast Consecration.
6. Cast Crusader Strike at 2 charges.
7. Cast Crusader Strike at 1 charge.
8. Cast Arcane Torrent (if Blood Elf).

To spend your Holy Power, use the following priority.

1. Cast Inquisition with at least 2 Holy Power.
2. Cast Execution Sentence.
3. Cast Templar's Verdict.
4. Cast Divine Storm (Empyrean Power).
It is important to note that, when generating Holy Power, if you have only 1 Crusader Strike charge available, you should immediately stop generating Holy Power and go to the Holy Power-spending priority. Stop immediately if you have 2 Crusader Strike charges available and you are using the ammer of Wrath talent, and one of the following two conditions is true:

1. Avenging Wrath is active;
2. Your target has less than 20% Health.

Always cast Avenging Wrath on cooldown.

--]]
-- -----------------------------------------------------------------------------
function HoJ:CheckDPSRotation_Ret_8dot1_Internal( skillsList )
	HoJ._useHolyPower = false;

	if ( not HoJ._showRotationFrame ) then
		return false;
	end

	if ( not HoJ._inCombat and not HoJ._targetCanAttack ) then
		-- HoJ:DebugMsg( "not in combat or target can't attack" );
		return false;
	end

	-- check if we should renew inquisition
	if ( HoJ._holyPower >= HoJ._minHolyPowerForInquisition and HoJ._inqState <= HOJ_INQ_NEARLY_EXPIRED ) then
		local _, _, _, inquisitionSelected, inquisitionAvailable = GetTalentInfo( 7, 3, 1 );
		if ( inquisitionAvailable and inquisitionSelected ) then
			--HoJ:DebugMsg( "Renew Inquisition" );	
			HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.inquisition, "1" );
			return true;
		end
	end

	-- if empyrean power buff is up and the target is in range, always use it first
	if ( HoJ._targetInRange[HoJ.SpellInfo.empyreanPower.range] ) then
		local epStart, epDuration, epRemaining = HoJ:GetBuffTimeRemainingWithoutGCD( HoJ.SpellInfo.empyreanPower.name );
		if ( epRemaining > 0.05 ) then
			--HoJ:DebugMsg( "Use Empyrean Power" );	
			HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.empyreanPower, "2" );
			return true;
		end
	end

	-- check for the next skill that is off cooldown AND won't overflow our holy power
	local nextHoPoGenerator = HoJ:GetNextHoPoGenerator( HoJ._holyPower );
	if ( nextHoPoGenerator == nil ) then
		-- we didn't find a skill to use, so we should try to use a holy power skill
		local holyPowerSkill = HoJ:ChooseHolyPowerSkill( skillsList );
		if ( holyPowerSkill ~= nil ) then
			--HoJ:DebugMsg( "Use Holy Power" );
			HoJ:AddIconDebug( skillsList, holyPowerSkill, "3" );
			HoJ._useHolyPower = true;
			return true;
		end

		return false;
	else
		--HoJ:DebugMsg( "Use generator " .. nextHoPoGenerator.name );	
		HoJ:AddIconDebug( skillsList, nextHoPoGenerator, "4" );
		return true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckDPSRotation_Ret_8dot1
-- -----------------------------------------------------------------------------
function HoJ:CheckDPSRotation_Ret_8dot1( skillsList )
	if ( HoJ:CheckDPSRotation_Ret_8dot1_Internal( skillsList ) ) then
		-- HoJ:DebugMsg( "Enabling Skills Button" );
		-- HoJ_RotationFrame:Enable();
		-- HoJ_RotationFrame:Show();
		-- HoJ_RotationFrame:EnableMouse( true );
	else
		-- HoJ:DebugMsg( "Disabling Skills Button" );
		-- HoJ_RotationFrame:Disable();
		-- HoJ_RotationFrame:Hide();
		-- HoJ_RotationFrame:EnableMouse( false );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckTankRotation_8_dot_1
-- -----------------------------------------------------------------------------
function HoJ:CheckTankRotation_8_dot_1( skillsList )
	if ( not HoJ._showRotationFrame ) then
		return;
	end

	if ( not HoJ._inCombat and not HoJ._targetCanAttack ) then
		return;
	end

	local hasConsecrationBuff = HoJ:FindBuff( HoJ.SpellInfo.consecration.name );
	local consecrationCdRemaining = HoJ:GetCDRemaining( HoJ.SpellInfo.consecration.id );
	--local timeSinceConsecrationCast = HoJ._curTime - HoJ.SpellInfo.consecration.lastCastTime;

	if ( not hasConsecrationBuff and consecrationCdRemaining <= 0 ) then
		HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.consecration, "2" );
		return true;
	end

	if ( HoJ:IsSkillUsable( HoJ.SpellInfo.judgment, HoJ.SKILL_USE_PREDICTION_TIME ) ) then
		HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.judgment, "1" );
		return true;
	end

	if ( HoJ:IsSkillUsable( HoJ.SpellInfo.avengersShield, HoJ.SKILL_USE_PREDICTION_TIME ) ) then
		HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.avengersShield, "3" );
		return true;
	end

	if ( HoJ:IsTalentSelected( HoJ.SpellInfo.blessedHammer ) ) then
		if ( HoJ:IsSkillUsable( HoJ.SpellInfo.blessedHammer, HoJ.SKILL_USE_PREDICTION_TIME ) ) then
			HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.blessedHammer, "4" );
			return true;
		end
	end

	if ( HoJ:IsSkillUsable( HoJ.SpellInfo.hammerOfTheRighteous, HoJ.SKILL_USE_PREDICTION_TIME ) ) then
		HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.hammerOfTheRighteous, "5" );
		return true;
	end	

	if ( consecrationCdRemaining <= 0 ) then
		HoJ:AddIconDebug( skillsList, HoJ.SpellInfo.consecration, "6" );
		return true;
	end

	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateSkill
-- -----------------------------------------------------------------------------
function HoJ:UpdateSkill( skillsList, skill, minCooldown, minHealthRatio, minHolyPower )
	minHealthRatio = minHealthRatio or 1.0;
	minHolyPower = minHolyPower or 0;
	
	if ( HoJ._holyPower < minHolyPower ) then
		HoJ:RemoveIcon( skillsList, skill );
		return;
	end
	
	local healthRatio = HoJ:GetHealthRatio( "player" );
	local cooldown = HoJ:GetCDRemaining( skill.id ) - HoJ._gcdRemaining;
	if ( cooldown < 0.00001 ) then
		cooldown = 0.0;
	end	
	skill.cooldownRemaining = cooldown;
	
	-- fixme make minCooldown and minHealthRatio member of spellInfo
	-- HoJ:DebugMsg( skill.name .. ", cd = " .. cooldown .. ", gcdr = " .. HoJ._gcdRemaining .. ", minCD = " .. minCooldown .. ", hr = " .. healthRatio .. ", minHr = " .. minHealthRatio );
	--if ( cooldown > 0.00001 and cooldown <= HoJ._gcdRemaining ) then
		-- always show skill that are on cooldown and if there's a gcd?
		--HoJ:AddIcon( skillsList, skill );
	if ( cooldown <= minCooldown and healthRatio <= minHealthRatio and HoJ._level >= skill.minLevel ) then
		HoJ:AddIcon( skillsList, skill, "" );
		--HoJ:DebugMsg( skill.name .. "enabled" );
	elseif ( HoJ._showSkillCDs and cooldown > 0.00001 ) then
		if ( cooldown <= 3.0 ) then
			HoJ:AddIcon( skillsList, skill, string.format( "%.1f", cooldown ) );
		else 
			HoJ:AddIcon( skillsList, skill, string.format( "%i", cooldown ) );
		end
	else
		HoJ:RemoveIcon( skillsList, skill );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckInterrupts
-- -----------------------------------------------------------------------------
function HoJ:CheckInterrupts( skillsList )
	if ( not self._showCooldownsFrame ) then
		return;
	end
	
	if ( not self._inCombat and not self._targetCanAttack ) then
		return;
	end
	
	if ( self._targetInRange[self.RANGE_10_YARDS] ) then
		if ( self:IsTalentSelected( self.SpellInfo.blindingLight ) and self:IsOffCooldown( self.SpellInfo.blindingLight, 0.0 ) ) then
			self:AddIcon( skillsList, self.SpellInfo.blindingLight )
		elseif ( HoJ:IsTalentSelected( self.SpellInfo.fistOfJustice ) and self:IsOffCooldown( self.SpellInfo.fistOfJustice, 0.0 ) ) then
			self:AddIcon( skillsList, self.SpellInfo.fistOfJustice )
		elseif ( self:IsOffCooldown( self.SpellInfo.hammerOfJustice, 0.0 ) ) then
			self:AddIcon( skillsList, self.SpellInfo.hammerOfJustice )
		end
	end
	
	if ( self._targetCanBeInterrupted and self._targetInRange[HoJ.RANGE_MELEE] and HoJ:IsOffCooldown( self.SpellInfo.rebuke, 0.0 ) ) then
		self:AddIcon( skillsList, self.SpellInfo.rebuke )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:ADDON_LOADED( ... )
-- -----------------------------------------------------------------------------
function HoJ._events:ADDON_LOADED( arg1 )
	if ( arg1 == "HammerOfJustas" ) then
		HoJ_OnAddonLoaded( self );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_ENTERING_WORLD
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_ENTERING_WORLD( ... )
	-- TODO: multiple greetings for the heck of it
	HoJ:Print( "Welcome, " .. HoJ._playerName .. ". May the light favor your hammer." );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_TARGET_CHANGED
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_TARGET_CHANGED( ... )
	--HoJ:UpdateSacredShieldIcon();
	-- HoJ:DebugMsg( "PLAYER_TARGET_CHANGED" );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_THREAT_LIST_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_THREAT_LIST_UPDATE( ... )
	--HoJ:DebugMsg( "UNIT_THREAT_LIST_UPDATE" );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_ENTER_COMBAT
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_REGEN_DISABLED( ... )
	HoJ:EnterCombat( ... );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LEAVE_COMBAT
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_REGEN_ENABLED( ... )
	HoJ:LeaveCombat( ... );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_TALENT_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_TALENT_UPDATE( ... )
	-- HoJ:DebugMsg( "PLAYER_TALENT_UPDATE" );
	-- HoJ._needsReset = true;
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:HOLY_POWER
-- -----------------------------------------------------------------------------
function HoJ._events:HOLY_POWER(...)
	HoJ._holyPower = UnitPower( "player", 9 );
	--HoJ:DebugMsg( HoJ._playerName .. " has " .. HoJ._holyPower .. " holy power." );
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:UNIT_SPELLCAST_SUCCEEDED
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_SPELLCAST_SUCCEEDED( unitID, spell, rank, lineID, spellID )
	if ( unitID == "player" ) then
		--HoJ:DebugMsg( "player UNIT_SPELLCAST_SUCCEEDED for " .. spell );
		--HoJ:DebugMsg( "spellID = " .. spellID .. ", inquisition.id = " .. HoJ.SpellInfo.inquisition.id );

		if ( spellID == HoJ.SpellInfo.inquisition.id and HoJ._inCombat ) then	
			HoJ:StartInquisitionCooldown();
		elseif ( spellID == HoJ.SpellInfo.consecration.id ) then
			HoJ.SpellInfo.consecration.lastCastTime = GetTime();
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_HEALTH
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_HEALTH( unit )
--[[	
	if ( unit == "player" ) then
		HoJ:DebugMsg( "Player damaged" )
		HoJ._showHealthTime = HoJ._curTime + 0.5
	end
]]	
end

-- -----------------------------------------------------------------------------
-- HoJ:SetHPAlphas
-- -----------------------------------------------------------------------------
function HoJ:SetHPAlphas( self, alpha, numVisible ) 
	local holyPowerAlpha = alpha;
	local templarsVerdictAlpha = alpha;

	if ( not HoJ._showHolyPowerFrame ) then
		holyPowerAlpha = 0.0;
	end
	if ( not HoJ._showTemplarsVerdictFrame ) then
		templarsVerdictAlpha = 0.0;
	end

	for i = 0, numVisible - 1 do
		--HoJ:DebugMsg( "Setting alpha " .. i );
		HoJ._HoPo[i]:SetAlpha( holyPowerAlpha );
		HoJ._HoPoGlow[i]:SetAlpha( holyPowerAlpha );
	end
	
	for i = numVisible, 4 do
		--HoJ:DebugMsg( "Clearing alpha " .. i );
		HoJ._HoPo[i]:SetAlpha( 0.0 );
		HoJ._HoPoGlow[i]:SetAlpha( 0.0 );
	end

	-- set alpha for big graphic
	if ( numVisible < 3 ) then -- or not HoJ._useHolyPower ) then
		HoJFrame.HPMaxed:SetAlpha( 0.0 );
		HoJFrame.HPMaxedGlow:SetAlpha( 0.0 );
	else
		HoJFrame.HPMaxed:SetAlpha( templarsVerdictAlpha );
		HoJFrame.HPMaxedGlow:SetAlpha( templarsVerdictAlpha );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateHealthHistory
-- -----------------------------------------------------------------------------
function HoJ:UpdateHealthHistory()
	if ( self._curTime >= self._nextHealthTime ) then
		if ( self._healthBuffer:isFull() ) then
			local h = self._healthBuffer:remove( 0 )
			if ( h == nil ) then
				self:DebugMsg( "ERROR! removed nil health value!" )
			end
		end

		self._healthBuffer:add( { time = HoJ._curTime, health = UnitHealth( "player" ) } )
		self._nextHealthTime = ( 1.0 / HoJ.HEALTH_SAMPLES_PER_SEC ) + self._curTime
		-- self:DebugMsg( "_nextHealthTime = " .. self._nextHealthTime )

		-- local show = self._curTime < self._showHealthTime

		local count = self._healthBuffer:count() - 1
		local size = self._healthBuffer.maxSize -- maximum history samples		
		local scale = count / size

		local tail = self._healthBuffer.tail
		local index = tail % size
		local lastHealth = self._healthBuffer[index].health
		local damage = 0
		local heals = 0
		local largestDamage = 0
		local futureHealth = HoJ._playerHealth
		local futureHealthTime = 0.0
		self._predictedDeathTime = 0.0
		self._predictedNearDeathTime = 0.0
		for i = 1, count do
			index = ( tail + i ) % size
			local h = self._healthBuffer[index].health

			local delta = h - lastHealth
			if ( delta ~= 0 ) then
				if ( delta < 0 ) then
					damage = damage - delta
					if ( delta < largestDamage ) then
						largestDamage = delta
					end
				else
					heals = heals + delta
				end

				futureHealth = futureHealth + delta
				futureHealthTime = i * ( HoJ.HEALTH_SAMPLE_SECONDS / size )
				if ( self._predictedNearDeathTime == 0.0 and futureHealth <= self._playerMaxHealth * 0.1 ) then
					self._predictedNearDeathTime = futureHealthTime
				elseif ( self._predictedDeathTime == 0.0 and futureHealth <= 0 ) then
					self._predictedDeathTime = futureHealthTime
				end
			end

			lastHealth = h
--[[
			if ( show ) then
				self:DebugMsg( "[" .. i .. "]: { health = " .. h .. " }" )
			end
]]			
		end

		self._damagePerSec = damage * scale
		self._healsPerSec = heals * scale 
		self._damagePerSecRatio = self._damagePerSec / self._playerMaxHealth
		self._healsPerSecRatio = self._healsPerSec / self._playerMaxHealth
		self._largestRecentDamageRatio = largestDamage / self._playerMaxHealth
		self._predictedHealth = futureHealth
		self._predictedHealthTime = futureHealthTime

--		if ( self._predictedHealthTime > 0 and self._predictedHealth < self._playerHealth ) then
--			self:DebugMsg( "FutureHealth = " .. self._predictedHealth .. " @ " .. self._predictedHealthTime )
--		end

--		if ( damage > 0 or heals > 0 ) then
--			self:DebugMsg( "Incoming DPS = " .. damage .. ", HPS = " .. heals )
--		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateHealth
-- -----------------------------------------------------------------------------
function HoJ:UpdateHealth()
	HoJ._playerHealth = UnitHealth( "player" )
	HoJ._playerMaxHealth = UnitHealthMax( "player" )
	
	if ( HoJ._playerMaxHealth == nil or HoJ._playerMaxHealth <= 0 ) then
		HoJ._playerMaxHealth = 1
	end

	HoJ._playerHealthRatio = HoJ._playerHealth / HoJ._playerMaxHealth

	HoJ:UpdateHealthHistory()
--[[
	if ( self._damagePerSecRatio > 0.1 ) then
		self:DebugMsg( "Taking damage >= " .. self._damagePerSecRatio * 100 .. "%" )
	end	
	if ( self._damagePerSecRatio > self._playerHealthRatio ) then
		self:DebugMsg( HoJ.STR_RED .. "Incoming DPS will kill in 1 second!!" )
	end
	if ( self._damagePerSecRatio * 2.0 > self._playerHealthRatio ) then
		self:DebugMsg( HoJ.STR_ORANGE .. "Incoming DPS will kill in 2 seconds!!" )
	end
	if ( self._damagePerSecRatio * 3.0 > self._playerHealthRatio ) then
		self:DebugMsg( HoJ.STR_YELLOW .. "Incoming DPS will kill in 3 seconds!!" )		
	end

	if ( self._predictedHealth <= 0 and self._predictedHealthTime > 0.0 ) then
		self:DebugMsg( HoJ.STR_RED .. "Predicted death in " .. self._predictedHealthTime .. " seconds!" )
	end

	if ( self._predictedDeathTime > 0.0 ) then
		self:DebugMsg( HoJ.STR_RED .. "Predicted death in " .. self._predictedDeathTime .. " seconds!" )
	end
	if ( self._predictedNearDeathTime > 0.0 ) then
		self:DebugMsg( HoJ.STR_ORANGE .. "Predicted near death in " .. self._predictedNearDeathTime .. " seconds!" )
	end	
]]	
end

-- -----------------------------------------------------------------------------
-- HoJ:InitHealth
-- -----------------------------------------------------------------------------
function HoJ:InitHealth()
	-- CircularBuffer:test()
	self._healthBuffer = CircularBuffer:new( HoJ.HEALTH_SAMPLES_PER_SEC * HoJ.HEALTH_SAMPLE_SECONDS );
--[[
	self._healthBuffer:add( { time = GetTime(), health = UnitHealth( "player" ) } )
	local h = self._healthBuffer:remove()
	if ( h ~= nil ) then
		self:DebugMsg( "h.time = " .. h.time )
		self:DebugMsg( "h.health = " .. h.health )
	else
		self:DebugMsg( "h == nil!" )
	end
]]
end

-- -----------------------------------------------------------------------------
-- HoJ:MouseOver
-- -----------------------------------------------------------------------------
function HoJ:MouseOver()
	-- At some point in Cata or Mists, EnableMouse became protected in combat.
	--if ( HoJ._inCombat ) then
	--	return;
	--end
	
	local mouseFocus = GetMouseFocus();
	for i, v in pairs( HoJ._windowFrames ) do
		if ( MouseIsOver( v ) ) then
			if ( IsControlKeyDown() ) then
				-- HoJ:DebugMsg( "Enabling mouse for frame '" .. v:GetName() .. "'" );
				v:EnableMouse( true );

				if ( IsMouseButtonDown( "LeftButton" ) ) then
					-- HoJ:DebugMsg( "Start moving frame '" .. v:GetName() .. "'" );
					v:StartMoving();
				end
			end
			if ( not IsMouseButtonDown( "LeftButton" ) ) then 
				v:StopMovingOrSizing();
			end
		else
			--HoJ:DebugMsg( "Disabling mouse for frame '" .. v:GetName() .. "'" );
			v:EnableMouse( false );
			v:StopMovingOrSizing();
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:EnableFrame
-- -----------------------------------------------------------------------------
function HoJ:EnableFrame( self )
	if ( HoJ._isPaladin ) then
		table.insert( HoJ._windowFrames, self );	
		self:Show();
		self:SetMovable( true );
		self:EnableMouse( true );
		self:SetPoint( "CENTER", UIParent );
	else
		self:Hide();
	end	
end

-- -----------------------------------------------------------------------------
-- HoJ_OnEvent
-- -----------------------------------------------------------------------------
function HoJ_OnEvent( self, event, ... )
	--HoJ:DebugMsg( "event = " .. event );

	-- do nothing for non-paladins
	if ( HoJ._isPaladin == false ) then
		return;
	end
	-- call event functions
	HoJ._events[ event ](self, ... );
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetFrames
-- -----------------------------------------------------------------------------

function HoJ:ResetFrames( frameList )
	-- clear all icon frames
	for key, value in pairs( frameList ) do
		if ( value.texture ~= nil ) then
			value.texture:SetTexture( nil )
			value.texture:Hide()
		end
		if ( value.cooldown ~= nil ) then
			value.cooldown:Hide()
		end
		value.spellInfo = nil
		--value.text:Hide();
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Reset
-- -----------------------------------------------------------------------------
function HoJ:Reset() 
	-- HoJ:DebugMsg( "Reset" );

	HoJ_InitSpellInfo();

	HoJ:ResetFrames( HoJ._buffFrames );
	HoJ:ResetFrames( HoJ._cooldownFrames );
	HoJ:ResetFrames( HoJ._rotationFrames );	
	
	HoJ:ClearIconList( HoJ._buffs );
	HoJ:ClearIconList( HoJ._cooldowns );
	HoJ:ClearIconList( HoJ._rotationSkills );	

	HoJ._glowAlpha = 0.0;
	HoJ:SetHPAlphas( self, HoJ._glowAlpha, HoJ._holyPower );

	HoJ._needsReset = false;
end

-- -----------------------------------------------------------------------------
-- HoJ_Update
-- -----------------------------------------------------------------------------
function HoJ_Update( self )
	if ( HoJ._level < 10 ) then
		return;
	end

	HoJ:UpdateGlobalCooldown();

	--HoJ:DebugMsg( "HoJ_Update" );
	-- update holy power amount	
	local lastHolyPower = HoJ._holyPower;
	HoJ._holyPower = UnitPower( "player", 9 );	
	
	-- we don't display our own cooldown text at some points when OmniCC is loaded
	if ( HoJ._omniCC == false ) then
		if ( IsAddOnLoaded( "OmniCC" ) ) then
			HoJ._omniCC = true;		
			--HoJ:DebugMsg( "OmniCC is loaded. HammerOfJustas will hide OmniCC's inquisition timer." );
		else
			HoJ._omniCC = false;
		end	
	end

	if ( HoJ._enabledForRole ) then
		HoJ:UpdateHealth()		
	end

	local target = "target";

	HoJ._targetCanAttack = false;
	HoJ._targetCanBeInterrupted = false;
	HoJ._targetHealthRatio = 0.0;
	HoJ._targetHealth = 0;
	HoJ._targetLevel = 1;
	HoJ._targetClassification = "";
	HoJ._targetIsTough = false;

	HoJ._targetInRange[HoJ.RANGE_NONE] = true;
	HoJ._targetInRange[HoJ.RANGE_SELF] = true;
	HoJ._targetInRange[HoJ.RANGE_MELEE] = false;
	HoJ._targetInRange[HoJ.RANGE_10_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_12_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_20_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_30_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_40_YARDS] = false;

	HoJ._targetExists = UnitExists( target );

	-- cache target stuff if we have a valid target
	if ( HoJ._targetExists and HoJ._enabledForRole ) then
		HoJ._targetReaction = UnitReaction( target, "player" );
		if ( HoJ._targetReaction ~= nil
			 and not UnitIsDead( target ) 
			 and UnitExists( target ) 
			 and HoJ._targetReaction < HoJ.REACTION_THRESHOLD ) then	-- 5 if you want to show the skill bar for yellow mobs
			HoJ._targetCanAttack = true;
		else
			HoJ._targetCanAttack = false;
		end	

		if ( HoJ._targetCanAttack ) then
			local channelName, _, _, _, _, _, channelNotInterruptible = UnitChannelInfo( target )
			local castName, _, _, _, _, _, _, castNotInterruptible = UnitCastingInfo( target )
			if ( ( channelName ~= nil and not channelNotInterruptible )
				or ( castName ~= nil and not castNotInterruptible ) ) then
				HoJ._targetCanBeInterrupted = true;
			end
			HoJ._targetHealthRatio = HoJ:GetHealthRatio( target );
		end

		HoJ._targetHealth = UnitHealth( target );	
		HoJ._targetLevel = UnitLevel( target );
		HoJ._targetClassification = UnitClassification( target );
		
		-- target range finding
		HoJ._targetInRange[HoJ.RANGE_MELEE] = HoJ:InRange( HoJ.SpellInfo.rebuke, target );
		
		if ( HoJ:IsTalentSelected( HoJ.SpellInfo.fistOfJustice ) ) then
			HoJ._targetInRange[HoJ.RANGE_10_YARDS] = HoJ:InRange( HoJ.SpellInfo.fistOfJustice, target );
		else
			HoJ._targetInRange[HoJ.RANGE_10_YARDS] = HoJ:InRange( HoJ.SpellInfo.hammerOfJustice, target );
		end

		if ( HoJ._role == HOJ_ROLE_DPS ) then
			HoJ._targetInRange[HoJ.RANGE_12_YARDS] = HoJ:InRange( HoJ.SpellInfo.bladeOfJustice, target )
		else
			-- only ret has a 12 yard spell to check with
			HoJ._targetInRange[HoJ.RANGE_12_YARDS] = HoJ._targetInRange[HoJ.RANGE_10_YARS]
		end

		if ( HoJ:IsTalentSelected( HoJ.SpellInfo.executionSentence ) ) then
			HoJ._targetInRange[HoJ.RANGE_20_YARDS] = HoJ:InRange( HoJ.SpellInfo.executionSentence, target )
		elseif ( HoJ._role == HOJ_ROLE_DPS ) then
			-- we have no other offensive 20 yards spells to test, so we use execution sentence as a stand-in
			HoJ._targetInRange[HoJ.RANGE_20_YARDS] = HoJ:InRange( HoJ.SpellInfo.bladeOfJustice, target )
		else
			-- prot has no 20 yard offensive spells
			HoJ._targetInRange[HoJ.RANGE_20_YARDS] = HoJ._targetInRange[HoJ.RANGE_10_YARDS]
		end

		HoJ._targetInRange[HoJ.RANGE_30_YARDS] = HoJ:InRange( HoJ.SpellInfo.judgment, target );

		-- note LoH or any non-aggressive spell will simply fail for non-enemy targets, so we just
		-- set it to 30, since if you're within 30 you're within 40.
		if ( HoJ._targetReaction == nil or HoJ._targetReaction >= HoJ.REACTION_THRESHOLD ) then
			HoJ._targetInRange[HoJ.RANGE_40_YARDS] = HoJ._targetInRange[HoJ.RANGE_30_YARDS];
		else
			HoJ._targetInRange[HoJ.RANGE_40_YARDS] = HoJ:InRange( HoJ.SpellInfo.layOnHands, target );
		end
		--HoJ:DebugMsg( "40 yards = " .. HoJ:BoolToStr( HoJ._targetInRange[HoJ.RANGE_40_YARDS] ) );

		-- target toughtness rating
		HoJ._targetIsTough = HoJ._targetClassification == "worldboss" 
							 or HoJ._targetClassification == "rareelite" 
							 or HoJ._targetClassification == "elite"
							 or HoJ._targetLevel > HoJ._level + 3
							 or HoJ._targetHealth / HoJ._playerHealth > 0.8;	
	end	

	local updateCooldowns = HoJ._updateType == HoJ.UPDATE_COOLDOWNS -- HoJ._curTime >= HoJ._updateCooldownsTime;	
	local updateBuffs = HoJ._updateType == HoJ.UPDATE_BUFFS -- HoJ._curTime >= HoJ._updateBuffsTime;
	local updateRotation = HoJ._updateType == HoJ.UPDATE_ROTATION

	-- rotate through the update types
	HoJ._updateType = HoJ._updateType + 1
	if ( HoJ._updateType > HoJ.UPDATE_MAX ) then
		HoJ._updateType = HoJ.UPDATE_START
	end

	-- sacred shield updates even if not in combat because we want to put it on
	-- before entering combat if we're targetting a hostile.
	local sortCooldowns = true;
	local sortRotation = false;
	if ( HoJ._role == HOJ_ROLE_TANK ) then
		if ( HoJ._enabledForRole ) then
			-- update health checks (incoming damage per second)
			-- HoJ:DebugMsg( "Updating ROLE_TANK" );

			-- HoJ:UpdateIncomingDPS();
			-- HoJ:CheckSeals();
			-- HoJ:CheckRighteousFury();
			-- WODONLY
			if ( updateCooldowns ) then
				HoJ:ClearIconList( HoJ._cooldowns );

				HoJ:CheckMitigationSkills();
				HoJ:CheckInterrupts( HoJ._cooldowns );

				HoJ:UpdateFrameIcons( HoJFrame_Cooldowns, HoJ._cooldowns, HoJ._cooldownFrames, sortCooldowns );
			elseif ( updateRotation ) then
				HoJ:ClearIconList( HoJ._rotationSkills );

				-- HoJ:CheckConsecration();
				HoJ:CheckTankRotation_8_dot_1( HoJ._rotationSkills );

				HoJ:UpdateFrameIcons( HoJFrame_Rotation, HoJ._rotationSkills, HoJ._rotationFrames, sortRotation );
			end
		end
	elseif ( HoJ._role == HOJ_ROLE_DPS ) then
		-- HoJ:DebugMsg( "Updating ROLE_DPS" );

		if ( updateBuffs ) then
			HoJ:ClearIconList( HoJ._buffs );

			HoJ:CheckProcdBuffs();
			HoJ:CheckBlessings();			
			-- HoJ:CheckSeals();
			HoJ:UpdateFrameIcons( HoJFrame_Buffs, HoJ._buffs, HoJ._buffFrames, true );
		end

		-- update cooldowns
		if ( updateCooldowns ) then
			HoJ:ClearIconList( HoJ._cooldowns );

			HoJ:CheckSelflessHealer();
			HoJ:CheckShieldOfVengeance();
			HoJ:CheckInquisition();
			HoJ:CheckAvengingWrath();
			HoJ:CheckInterrupts( HoJ._cooldowns );

			HoJ:UpdateFrameIcons( HoJFrame_Cooldowns, HoJ._cooldowns, HoJ._cooldownFrames, sortCooldowns );
		end

		-- update rotation
		if ( updateRotation ) then
			HoJ:ClearIconList( HoJ._rotationSkills );

			-- This is not working and may be causing issues.
			-- HoJ:CheckRetSkills( HoJ._cooldowns );
			HoJ:CheckDPSRotation_Ret_8dot1( HoJ._rotationSkills );

			HoJ:UpdateFrameIcons( HoJFrame_Rotation, HoJ._rotationSkills, HoJ._rotationFrames, sortRotation );
		end
	elseif ( HoJ._role == HOJ_ROLE_HEALING ) then
		if ( HoJ._enabledForRole ) then
			--HoJ:DebugMsg( "Updating ROLE_HEALING" );
		end
	else
		--HoJ:DebugMsg( "WE HAVE NO IDEA WHAT OUR ROLE IS!" );
	end

	if ( not HoJ._enabledForRole ) then
		HoJ:ResetFrames( HoJ._buffFrames );
		HoJ:ResetFrames( HoJ._cooldownFrames );
		HoJ:ResetFrames( HoJ._rotationFrames );
	end
--[[
	if ( updateBuffs ) then
		HoJ._updateBuffsTime = HoJ._curTime + 0.25;
	end
	if ( updateCooldowns ) then
		HoJ._updateCooldownsTime = HoJ._curTime + 0.25;
	end
]]
end		

-- -----------------------------------------------------------------------------
-- HoJ_UpdateGlow
-- -----------------------------------------------------------------------------
function HoJ_UpdateGlow( self, elapsed )
	if ( HoJ._glowAlpha > HoJ._maxGlowAlpha ) then
		HoJ._glowScale = 1.0 - ( 4.0 * elapsed );
	elseif ( HoJ._glowAlpha < HoJ._minGlowAlpha ) then
		HoJ._glowScale = 1.0 + ( 4.0 * elapsed );
	end

	HoJ._glowAlpha = HoJ._glowAlpha * HoJ._glowScale;
	if ( HoJ._glowAlpha <= 0.0 ) then
		HoJ._glowAlpha = HoJ._minGlowAlpha;
	elseif ( HoJ._glowAlpha > 1.0 ) then
		HoJ._glowAlpha = HoJ._maxGlowAlpha;
	end
	--HoJ:DebugMsg( "_glowAlpha = " .. HoJ._glowAlpha );
	--HoJ:DebugMsg( "_glowScale = " .. HoJ._glowScale );
	
	-- update holy power icons
	if ( lastHolyPower ~= 3 and HoJ._holyPower == 3 ) then
		--PlaySound( "LEVELUP" );
		--PlaySound( "igPVPUpdate" );
		-- BFAFIX
		-- HoJ:DebugMsg( "Playing detected sound" );
		-- PlaySoundFile( "Interface\\AddOns\\HammerOfJustas\\Sounds\\detected-nearby.wav" );
	end
	
	finalAlpha = HoJ._glowAlpha;
	if ( IsMounted() ) then
		finalAlpha = finalAlpha * 0.1875;
	end
	
	HoJ:SetHPAlphas( self, finalAlpha, HoJ._holyPower );

	HoJ_RotationFrame:SetAlpha( 0.666 );		
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateGlobalCooldown
-- -----------------------------------------------------------------------------
function HoJ:UpdateGlobalCooldown()
	local start, duration, enabled = GetSpellCooldown( HoJ.SpellInfo.globalCooldown.id );
	if ( start > 0 and duration > 0 ) then
		HoJ._gcdDuration = duration;
		HoJ._gcdRemaining = ( start + duration ) - HoJ._curTime;
	else
		HoJ._gcdRemaining = 0;
	end

	--HoJ:DebugMsg( "Global Cooldown: duration = " .. HoJ._gcdDuration .. ", remaining = " .. HoJ._gcdRemaining );
end

-- -----------------------------------------------------------------------------
-- HoJ_UpdateRender
-- -----------------------------------------------------------------------------
function HoJ_UpdateRender( self, elapsed )
	if ( HoJ._enabledForRole ) then
		HoJ_UpdateGlow( self, elapsed );
	else
		HoJ_RotationFrame:SetAlpha( 0.0 );
	end
	
	-- inquisition stuff
	HoJ:UpdateInquisitionFrames( self );
end

-- -----------------------------------------------------------------------------
-- HoJ_OnUpdate
-- -----------------------------------------------------------------------------
function HoJ_OnUpdate( self, elapsed ) 
	HoJ._frameNum = HoJ._frameNum + 1;

	if ( not HoJ._enabled ) then
		if ( HoJ._needsReset ) then
			HoJ:Reset();
		end
		return;
	end
	
	if ( HoJ._enableProfiling ) then
		ResetCPUUsage();
	end

	if ( HoJ._playerRaceId == 0 ) then
		_, _, HoJ._playerRaceId = UnitRace( "player" )
	end
	
	local lastRole = HoJ._role;
	HoJ:UpdateRole();
	if ( HoJ._role ~= lastRole ) then
		HoJ:Print( "HammerOfJustas: role changed to " .. HoJ.HOJ_ROLE_NAMES[ HoJ._role ] .. "." );
		HoJ._needsReset = true;
	end

	if ( HoJ._needsReset ) then
		HoJ:Reset();
	end

	if ( HoJ._role == HOJ_ROLE_HEALING ) then
		if ( HoJ._enabledForRole ) then
			HoJ:Print( "Hammer of Justas: disabled for Holy." );
		end
		HoJ._enabledForRole = false;
	elseif ( HoJ._role == HOJ_ROLE_TANK ) then
		if ( not HoJ._enabledForRole ) then
			HoJ:Print( "Hammer of Justas: enabled for Protection." );
		end
		HoJ._enabledForRole = true;
	else
		if ( not HoJ._enabledForRole ) then
			HoJ:Print( "Hammer of Justas: enabled for Retribution." );		
		end
		HoJ._enabledForRole = true;
	end

	--HoJ:DebugMsg( "HoJ_OnUpdate" );
	-- do nothing for non-paladins
	if ( HoJ._isPaladin == false ) then
		return;
	end	

	-- update time
	HoJ._curTime = GetTime();
	HoJ:MouseOver();
	
	-- update general logic
	if ( HoJ._curTime >= HoJ._nextUpdateTime ) then
		HoJ._nextUpdateTime = HoJ._curTime + HoJ.UPDATE_INTERVAL
		HoJ_Update( self )
	end
--[[
	HoJ._updateElapsedTime = HoJ._updateElapsedTime + elapsed;
	if ( HoJ._updateElapsedTime > HoJ.UPDATE_INTERVAL ) then
		update = true;
		HoJ._updateElapsedTime = HoJ._updateElapsedTime - HoJ.UPDATE_INTERVAL;
	end
]]

	-- update render logic
	if ( HoJ._curTime >= HoJ._nextRenderTime ) then
		HoJ._nextRenderTime = HoJ._curTime + HoJ.RENDER_INTERVAL;
		HoJ_UpdateRender( self, elapsed )
	end
--[[	
	HoJ._renderElapsedTime = HoJ._updateElapsedTime + elapsed;	
	if ( HoJ._renderElapsedTime > HoJ.RENDER_INTERVAL ) then
		updateRender = true;
		HoJ._renderElapsedTime = HoJ._renderElapsedTime - HoJ.RENDER_INTERVAL;
	end
]]

	if ( HoJ._enableProfiling ) then	
		UpdateAddOnCPUUsage();
		local addonMs = GetAddOnCPUUsage( "HammerOfJustas" );
		if ( addonMs ~= nil and addonMs > HoJ._profileThreshold ) then
			HoJ:Msg( "CPU: " .. addonMs .. "ms" );
		end
	end
end

function HoJ:FindChild( self, substr )
	--HoJ:DebugMsg( "FindChild( " .. self:GetName() .. " )" );
	local children = self:GetChildren();
	for i, child in ipairs( children ) do
		HoJ:DebugMsg( "Child " .. i .. " is " .. child.GetName() );
		if ( string.find( child:GetName(), substr ) ~= nil ) then
			return child;
		end
	end
	return nil;
end

function HoJ:FindRegion( self, substr )
	--HoJ:DebugMsg( "FindRegion( " .. self:GetName() .. " )" );
	local regions = { self:GetRegions() }
	for i, region in ipairs( regions ) do
		-- HoJ:DebugMsg( "Region " .. i .. " is " .. region.GetName() );
		-- if ( string.find( region:GetName(), substr ) ~= nil ) then
		if ( region and region:GetObjectType() == "FontString" ) then
			return region;
		end
	end
	return nil;
end

function HoJ:SetTitleTextColor( self, colorR, colorG, colorB, colorA )
	--HoJ:DebugMsg( "SetTitleTextColor( " .. self:GetName() .. " )" );
	local child = HoJ:FindChild( self, "TitleText" );
	if ( child ~= nil ) then
		child:SetTextColor( colorR, colorG, colorB, colorA );
	end
end

function HoJ:ShowTitle( self )
	local region = HoJ:FindRegion( self, "TitleText", self:GetRegions() );
	if ( region ~= nil ) then
		region:Show();
	end
end

function HoJ:HideTitle( self )
	local region = HoJ:FindRegion( self, "TitleText", self:GetRegions() );
	if ( region ~= nil ) then
		region:Hide();
	end
end

function HoJ:IsHoJFrame( self )
	for i, v in ipairs( HoJ._windowFrames ) do
		if ( v == self ) then
			return true;
		end
	end			
	return false;
end	

-- -----------------------------------------------------------------------------
-- HoJ_OnEnter
-- -----------------------------------------------------------------------------
function HoJ_OnEnter( self )
	--HoJ:DebugMsg( "HoJ_OnEnter( " .. self:GetName() .. " )" );
	if ( not HoJ._enabledForRole ) then
		return;
	end

	if ( IsControlKeyDown() ) then
		-- if this isn't one of the parent windows, recurse until we get to one
		-- then, turn on background alpha for all the windows.
		if ( not HoJ:IsHoJFrame( self ) ) then
			HoJ_OnEnter( self:GetParent() );
		else
			for i, wf in ipairs( HoJ._windowFrames ) do
				wf:SetBackdropColor( 0.5, 0.5, 0.5, 0.5 );
				--HoJ:SetTitleTextColor( wf, 0.5, 0.5, 0.5, 0.5 );
				HoJ:ShowTitle( wf );
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLeave
-- -----------------------------------------------------------------------------
function HoJ_OnLeave( self )
	--HoJ:DebugMsg( "HoJ_OnLeave( " .. self:GetName() .. " )" );
	if ( not HoJ._enabledForRole ) then
		return;
	end

	-- if this isn't one of the parent windows, recurse until we get to one
	-- then, turn on background alpha for all the windows.
	if ( not HoJ:IsHoJFrame( self ) ) then
		HoJ_OnLeave( self:GetParent() );
	else
		for i, wf in ipairs( HoJ._windowFrames ) do
			wf:SetBackdropColor( 0.5, 0.5, 0.5, 0.0 );	
			--HoJ:SetTitleTextColor( wf, 0.5, 0.5, 0.5, 0.0 );
			HoJ:HideTitle( wf );
		end
	end

	-- self:StopMovingOrSizing();
end

-- -----------------------------------------------------------------------------
-- HoJ_OnDragStart
-- -----------------------------------------------------------------------------
function HoJ_OnDragStart( self, button )
	-- HoJ:DebugMsg( "HoJ_OnDragStart down for '" .. self:GetName() .. "'" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnDragStop
-- -----------------------------------------------------------------------------
function HoJ_OnDragStop( self )
	-- HoJ:DebugMsg( "HoJ_OnDragStop for '" .. self:GetName() .. "'" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnMouseDown
-- -----------------------------------------------------------------------------
function HoJ_OnMouseDown( self, button )
	-- HoJ:DebugMsg( "HoJ_OnMouseDown for '" .. self:GetName() .. "'" )
	--if ( IsControlKeyDown() and not self.isLocked ) then
    -- 	self:StartMoving()
   	--end		
end

-- -----------------------------------------------------------------------------
-- HoJ_OnMouseUp
-- -----------------------------------------------------------------------------
function HoJ_OnMouseUp( self, button )
	-- HoJ:DebugMsg( "HoJ_OnMouseUp for '" .. self:GetName() .. "'" )
	--self:StopMovingOrSizing();
end

-- -----------------------------------------------------------------------------
-- HoJ_Filter_Whisper
-- -----------------------------------------------------------------------------
function HoJ_Filter_Whisper( self, event, msg, player, _, _, _, gm )
	local blocked = false;
	if ( HoJ._blockedNames[ player ] ~= nil ) then
		blocked = true;
	end

	if ( blocked == true ) then 
		-- HoJ:Print( type( HoJ._ignoreHistory[ player ] ) );
		if ( HoJ._ignoreHistory[ player ] == nil ) then
			HoJ._ignoreHistory[ player ] = { };
		end
		table.insert( HoJ._ignoreHistory[ player ], msg );
		if ( HoJ._silentIgnore == false ) then
			if ( HoJ._stealthIgnore == true ) then
				HoJ:Print( HoJ._stealthIgnoreMessage );
			else 
				HoJ:Print( "Whisper from " .. player .. " blocked." );
			end
		end
		return true;
	end
	-- do default behavior
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnOK
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnOK()
	HoJ:WriteSavedVariables();
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnCancel
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnCancel()
	--HoJ:DebugMsg( "AddonOptions_OnCancel" );
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnDefault
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnDefault()
	--HoJ:DebugMsg( "AddonOptions_OnDefault" );
end

-- -----------------------------------------------------------------------------
-- HoJ:AddonOptions_OnRefresh
-- -----------------------------------------------------------------------------
function HoJ:AddonOptions_OnRefresh()
	--HoJ:DebugMsg( "AddonOptions_OnRefresh" );
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowSpellOnBar
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowSpellOnBar( spell ) 
	spell.showOnBar = not spell.showOnBar 
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowSkillSFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowSkillSFrame() 
	if ( HoJ._showCooldownsFrame ) then
		--HoJ:DebugMsg( "_showCooldownsFrame = false" );
		HoJ._showCooldownsFrame = false;
	else
		--HoJ:DebugMsg( "_showCooldownsFrame = true" );
		HoJ._showCooldownsFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowBuffsFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowBuffsFrame() 
	if ( HoJ._showBuffsFrame ) then
		--HoJ:DebugMsg( "_showBuffsFrame = false" );
		HoJ._showBuffsFrame = false;
	else
		--HoJ:DebugMsg( "_showBuffsFrame = true" );
		HoJ._showBuffsFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowRotationFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowRotationFrame()
	if ( HoJ._showRotationFrame ) then
		HoJ._showRotationFrame = false;
	else
		HoJ._showRotationFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:TogglehowTemplarsVerdictFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowTemplarsVerdictFrame() 
	if ( HoJ._showTemplarsVerdictFrame ) then
		HoJ._showTemplarsVerdictFrame = false;
	else
		HoJ._showTemplarsVerdictFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowHolyPowerFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowHolyPowerFrame() 
	if ( HoJ._showHolyPowerFrame ) then
		HoJ._showHolyPowerFrame = false;
	else
		HoJ._showHolyPowerFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ToggleShowInquisitionFrame
-- -----------------------------------------------------------------------------
function HoJ:ToggleShowInquisitionFrame() 
	if ( HoJ._showInquisitionFrame ) then
		HoJ._showInquisitionFrame = false;
	else
		HoJ._showInquisitionFrame = true;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupCheckBox
-- -----------------------------------------------------------------------------
function HoJ:SetupCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, checked, text )
	local cb = CreateFrame( "CheckButton", name, parentFrame, "OptionsCheckButtonTemplate" );
	cb:SetPoint( anchor1, anchorFrame, anchor2, xoffset, yoffset );
	--cb:SetScript( "OnClick", function ( self ) HoJ:ToggleShowSpellOnBar( spellInfo ) end );
	cb:SetChecked( checked );
	-- the following is really pretty stupid and was a PITA to figure out
	getglobal( cb:GetName() .. 'Text' ):SetText( text );
	return cb;
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupSpellCheckBox
-- -----------------------------------------------------------------------------
function HoJ:SetupSpellCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, spellInfo, text )
	local cb = CreateFrame( "CheckButton", name, parentFrame, "OptionsCheckButtonTemplate" );
	cb:SetPoint( anchor1, anchorFrame, anchor2, xoffset, yoffset );
	cb:SetScript( "OnClick", function ( self ) HoJ:ToggleShowSpellOnBar( spellInfo ) end );
	cb:SetChecked( spellInfo.showOnBar );
	-- the following is really pretty stupid and was a PITA to figure out
	if ( spellInfo.spellType == HoJ.SPELL_BUFF or spellInfo.spellType == HoJ.SPELL_TALENT_BUFF ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Buff)" );
	elseif ( spellInfo.spellType == HoJ.SPELL_SKILL ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Skill)" );	
	elseif ( spellInfo.spellType == HoJ.SPELL_TALENT ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Talent)" );			
	elseif ( spellInfo.spellType == HoJ.SPELL_SHAPESHIFT ) then
		getglobal( cb:GetName() .. 'Text' ):SetText( text .. " (Seal)" );			
	else
		getglobal( cb:GetName() .. 'Text' ):SetText( text );
	end
	return cb;
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupShowSpellCheckBox
-- -----------------------------------------------------------------------------
function HoJ:SetupShowSpellCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, spellInfo, count )
	return HoJ:SetupSpellCheckBox( parentFrame, anchorFrame, name, anchor1, anchor2, xoffset, yoffset, spellInfo, HOJ_SHOW .. spellInfo.name );
end

-- -----------------------------------------------------------------------------
-- HoJ:OnScrollValueChanged
-- -----------------------------------------------------------------------------
function HoJ:OnScrollValueChanged( slider, value ) 
	--HoJ:DebugMsg( "OnScrollValueChanged(): slider = " .. slider.name .. ", value = " .. value );
	slider.scrollView:SetVerticalScroll( value );
	slider.scrollView:UpdateScrollChildRect();
	local itemFrame = slider.scrollView:GetScrollChild();
	itemFrame:SetPoint( "TOPLEFT", slider.scrollView, "TOPLEFT", 0, value );
	itemFrame:SetPoint( "TOPRIGHT", slider.scrollView, "TOPRIGHT", 0, value );
end

-- -----------------------------------------------------------------------------
-- HoJ:scrollView_OnMouseWheel
-- -----------------------------------------------------------------------------
local function scrollView_OnMouseWheel( self, delta )
	local cur = self.slider:GetValue();
	local min, max = self.slider:GetMinMaxValues();
	local new = cur + ( delta * -1 * self.itemHeight );
	if ( new < min ) then
		new = min;
	elseif ( new > max ) then
		new = max;
	end
	self.slider:SetValue( new );
end

-- -----------------------------------------------------------------------------
-- HoJ:SetupAddOnOptions
-- -----------------------------------------------------------------------------
function HoJ:SetupAddOnOptions() 
	--HoJ:DebugMsg( "HoJ:SetupAddOnOptions" );
	local panel = CreateFrame( "FRAME", "HoJOptions" );
	--panel.name = "Hammer of Justas";
	panel.name = GetAddOnMetadata( "HammerOfJustas", "Title" );
	panel.parent = nil;
	panel.okay = function ( self ) HoJ:AddonOptions_OnOK( self ) end;
	panel.cancel = function ( self ) HoJ:AddonOptions_OnCancel( self ) end;
	panel.defalt = function ( self ) HoJ:AddonOptions_OnDefault( self ) end;
	panel.refresh = function ( self ) HoJ:AddonOptions_OnRefresh( self ) end;
	InterfaceOptions_AddCategory( panel );
	
	local panelHeader = panel:CreateFontString( nil, "ARTWORK" );
	panelHeader:SetFont( "Fonts\\FRIZQT__.TTF", 24 );
	panelHeader:SetJustifyH( "LEFT" );
	panelHeader:SetJustifyV( "TOP" );
	panelHeader:ClearAllPoints();
	panelHeader:SetPoint( "TOPLEFT", 16, -16 );
	panelHeader:SetText( "Hammer of Justas v. " .. HoJ:GetVersionString() );
	
	-- TODO: make this check list a class
	local checkListHeader = panel:CreateFontString( nil, "ARTWORK" );
	checkListHeader:SetFont( "Fonts\\FRIZQT__.TTF", 18 );
	checkListHeader:SetJustifyH( "LEFT" );
	checkListHeader:SetJustifyV( "TOP" );
	checkListHeader:ClearAllPoints();
	checkListHeader:SetTextColor( 1, 1, 1, 1 );
	checkListHeader:SetPoint( "TOPLEFT", panelHeader, "BOTTOMLEFT", 0, -16 );
	checkListHeader:SetText( "Skill Selection" );

	local checkListDescription = panel:CreateFontString( nil, "ARTWORK" );
	checkListDescription:SetFont( "Fonts\\FRIZQT__.TTF", 11 );
	checkListDescription:SetJustifyH( "LEFT" );
	checkListDescription:SetJustifyV( "TOP" );
	checkListDescription:ClearAllPoints();
	checkListDescription:SetTextColor( 1, 1, 1, 1 );
	checkListDescription:SetPoint( "TOPLEFT", panelHeader, "BOTTOMLEFT", 0, -40 );
	checkListDescription:SetText( 	"Display of individual cooldowns, talents and buffs can be toggled in the rotation, buffs and cooldowns frames.\n" ..
									"Note that turning off a skill in the rotation frame will likely result in an 'empty' rotation action\n" ..
									"when that skill is chosen for the rotation." );

	local SCROLL_BORDER_HEIGHT = 24 * 8 + EDGE_SIZE * 2;
	local scrollBorder = CreateFrame( "FRAME", "scrollBorder", panel );
	scrollBorder:SetPoint( "TOPLEFT", checkListDescription, "BOTTOMLEFT", 0, 0 );
	scrollBorder:SetPoint( "TOPRIGHT", panel, "TOPRIGHT", -16, 0 );
	scrollBorder:SetPoint( "BOTTOMLEFT", checkListDescription, "BOTTOMLEFT", 0, -SCROLL_BORDER_HEIGHT );
	scrollBorder:SetBackdrop( listBackdrop );

	local scrollView = CreateFrame( "SCROLLFRAME", "scrollView", scrollBorder );
	scrollView:EnableMouseWheel( true );
	scrollView:SetScript( "OnMouseWheel", scrollView_OnMouseWheel );
	scrollView:SetPoint( "TOPLEFT", scrollBorder, "TOPLEFT", EDGE_SIZE, -EDGE_SIZE );
	scrollView:SetPoint( "TOPRIGHT", scrollBorder, "TOPRIGHT", -EDGE_SIZE, -EDGE_SIZE );
	scrollView:SetPoint( "BOTTOMLEFT", scrollBorder, "BOTTOMLEFT", EDGE_SIZE,
	EDGE_SIZE );
	scrollView:SetPoint( "BOTTOMRIGHT", scrollBorder, "BOTTOMRIGHT", -EDGE_SIZE, EDGE_SIZE );
	
	local checkFrame = CreateFrame( "FRAME", "itemFrame", scrollView );
	scrollView:SetScrollChild( checkFrame );
	
	checkFrame:SetPoint( "TOPLEFT", scrollView, "TOPLEFT", 0, 0 );
	--checkFrame:SetPoint( "TOPRIGHT", scrollView, "TOPRIGHT", -EDGE_SIZE, -EDGE_SIZE * 2 );
	--checkFrame:SetPoint( "BOTTOMLEFT", scrollView, "BOTTOMLEFT", EDGE_SIZE, EDGE_SIZE );
	--checkFrame:SetPoint( "BOTTOMRIGHT", scrollView, "BOTTOMRIGHT", -EDGE_SIZE, EDGE_SIZE );
	checkFrame:Lower();

	local YPAD = 2;
	
	local prev = checkFrame;
	local anchor = "TOPLEFT";
	local count = 0;
	local spellIndex = 0;
	for key, value in pairs( HoJ.SpellInfo ) do
		if ( value.showCheckbox ) then
			--HoJ:DebugMsg( key .. " is " .. value.name );
			if ( key == nil ) then
				HoJ:DebugMsg( "got nil name for spellInfo id " .. value.id );
			elseif ( value == nil ) then
				HoJ:DebugMsg( "got nil value for spellInfo id " .. value.id );
			elseif ( value.name == nil ) then
				HoJ:DebugMsg( "got nil value.name for spellInfo id " .. value.id );
			else
				prev = HoJ:SetupShowSpellCheckBox( checkFrame, prev, key, "TOPLEFT", anchor, 0, YPAD, value, count );
				if ( count == 0 ) then
					anchor = "BOTTOMLEFT";
				end
			end
			count = count + 1;
		end
		spellIndex = spellIndex + 1;
	end

	-- There must be a better way to calculate the distance from the top of the first item to the bottom of the last item
	scrollView.itemHeight = prev:GetHeight() - YPAD;
	local itemsInScrollView = 4;
	local totalHeight = ( scrollView.itemHeight ) * ( count - itemsInScrollView );
	checkFrame:SetSize( 400, totalHeight ); 
	scrollView:SetSize( 400, totalHeight );
	
	local slider = CreateFrame( "SLIDER", "slider", scrollView );
	slider.name = "slider";
	slider:SetOrientation( "VERTICAL" );
	slider:SetHitRectInsets( 0, 0, -10, 0 );
	slider:SetBackdrop( sliderBackdrop );
	slider:SetWidth( SCROLLBAR_WIDTH );
	slider:SetThumbTexture( "Interface\\Buttons\\UI-SliderBar-Button-Vertical" );
	slider.scrollView = scrollView;
	slider.slider = slider;	
	slider:SetPoint( "TOPLEFT", scrollView, "TOPRIGHT", -SCROLLBAR_WIDTH, 0 );
	slider:SetPoint( "BOTTOMLEFT", scrollView, "BOTTOMRIGHT", -SCROLLBAR_WIDTH, 0 );
	slider:SetScript( "OnValueChanged", function ( slider, value ) HoJ:OnScrollValueChanged( slider, value ) end );
	slider:SetMinMaxValues( 0, totalHeight );
	slider:SetValueStep( 1 );
	slider:SetValue( 0 );
	scrollView.slider = slider;
	

	-- TODO: make this check list a class
	local frameToggleHeader = panel:CreateFontString( nil, "ARTWORK" );
	frameToggleHeader:SetFont( "Fonts\\FRIZQT__.TTF", 18 );
	frameToggleHeader:SetJustifyH( "LEFT" );
	frameToggleHeader:SetJustifyV( "TOP" );
	frameToggleHeader:ClearAllPoints();
	frameToggleHeader:SetTextColor( 1, 1, 1, 1 );
	frameToggleHeader:SetPoint( "TOPLEFT", scrollBorder, "BOTTOMLEFT", 0, -16 );
	frameToggleHeader:SetText( "Frame Toggles" );

	-- add check box for skills bar
	local checkShowCooldownsFrame = HoJ:SetupCheckBox( panel, frameToggleHeader, "checkCooldowns", "TOPLEFT", "BOTTOMLEFT", 0, 0, HoJ._showCooldownsFrame, "Show Cooldowns" );
	checkShowCooldownsFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowSkillSFrame() end );
	checkShowCooldownsFrame:Show();
	
	-- add check box for buffs bar
	local checkShowBuffsFrame = HoJ:SetupCheckBox( panel, checkShowCooldownsFrame, "checkBuffs", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showBuffsFrame, "Show Buffs" );
	checkShowBuffsFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowBuffsFrame() end );
	checkShowBuffsFrame:Show();

	local checkShowRotationFrame = HoJ:SetupCheckBox( panel, checkShowBuffsFrame, "checkRotation", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showRotationFrame, "Show Rotation" );
	checkShowRotationFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowRotationFrame() end );
	checkShowRotationFrame:Show();	

	local checkShowTemplarsVerdictFrame = HoJ:SetupCheckBox( panel, checkShowCooldownsFrame, "checkShowTemplarsVerdict", "TOPLEFT", "BOTTOMLEFT", 0, 0, HoJ._showTemplarsVerdictFrame, "Show Templars Verdict" );
	checkShowTemplarsVerdictFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowTemplarsVerdictFrame() end );
	checkShowTemplarsVerdictFrame:Show();		

	local checkShowHolyPowerFrame = HoJ:SetupCheckBox( panel, checkShowTemplarsVerdictFrame, "checkShowHolyPower", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showHolyPowerFrame, "Show Holy Power" );
	checkShowHolyPowerFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowHolyPowerFrame() end );
	checkShowHolyPowerFrame:Show();

	local checkShowInquisitionFrame = HoJ:SetupCheckBox( panel, checkShowHolyPowerFrame, "checkShowInquisition", "TOPLEFT", "TOPLEFT", 200, 0, HoJ._showHolyPowerFrame, "Show Inquisition" );
	checkShowInquisitionFrame:SetScript( "OnClick", function ( self ) HoJ:ToggleShowInquisitionFrame() end );
	checkShowInquisitionFrame:Show();				
	
	scrollView:Show();
	checkFrame:Show();
	slider:Show();
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad
-- -----------------------------------------------------------------------------
function HoJ_OnLoad( self )
	--HoJ:DebugMsg( "HoJ_OnLoad : " .. self:GetName() );
	HoJ._curTime = GetTime();

	HoJ:InitHealth();

	local _, className = UnitClass( "player" );
	if ( className == "PALADIN" ) then
		HoJ._isPaladin = true;
	else
		HoJ._isPaladin = false;
	end

	HoJ:EnableFrame( self );

	local _, className = UnitClass( "player" );
	if ( HoJ._isPaladin ~= true ) then
		HoJ:Print( "The Hammer of Justas addon is intended for paladins only. The addon is disabled." );
		return;
	end

	HoJ._level = UnitLevel( "player" );
	HoJ._playerName = UnitName( "player" );	
	
	-- variables loaded
	HoJFrame:RegisterEvent( "ADDON_LOADED" );

	-- holy power gained
	-- BFAFIX
	-- HoJFrame:RegisterEvent( "HOLY_POWER" );

	-- handle player entering the world		
	HoJFrame:RegisterEvent( "PLAYER_ENTERING_WORLD" )

	-- handle player leaving the world
	-- HoJFrame:RegisterEvent( "PLAYER_LEAVING_WORLD" )

	-- check for changing into a bg or arena, etc.
	-- HoJFrame:RegisterEvent( "ZONE_CHANGED_NEW_AREA" );
	-- check for an attackable target
	HoJFrame:RegisterEvent( "PLAYER_TARGET_CHANGED" );

	-- check for entering combat
	--HoJFrame:RegisterEvent( "UNIT_COMBAT" );
	--HoJFrame:RegisterEvent( "PLAYER_ENTER_COMBAT" );
	HoJFrame:RegisterEvent( "PLAYER_REGEN_DISABLED" );

	-- check for leaving combat
	--HoJFrame:RegisterEvent( "PLAYER_LEAVE_COMBAT" );
	HoJFrame:RegisterEvent( "PLAYER_REGEN_ENABLED" );

	-- HoJFrame:RegisterEvent( "PLAYER_TALENT_UPDATE" );

	-- check for successful spellcasts
	HoJFrame:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" );	

	-- check for combat?
	-- HoJFrame:RegisterEvent( "UNIT_COMBAT" );
	-- ???
	-- HoJFrame:RegisterEvent( "VARIABLES_LOADED" );
	-- check auras
	--HoJFrame:RegisterEvent( "UNIT_AURA" );
	-- check mana changes
	-- HoJFrame:RegisterEvent( "UNIT_MANA" );
	-- check health changes
	HoJFrame:RegisterEvent( "UNIT_HEALTH" );

	--HoJ:Print( HoJ.ADDON_NAME .. " version " .. HoJ:GetVersionString() .. " loaded." );
end

-- -----------------------------------------------------------------------------
-- HoJ_OnAddonLoad
-- -----------------------------------------------------------------------------
function HoJ_OnAddonLoaded( self )
	HoJ:Print( HoJ.ADDON_NAME .. " version " .. HoJ:GetVersionString() .. " loaded." );
	--HoJ:DebugMsg( "HoJ_OnAddonLoaded : " .. self:GetName() );

	HoJ:InitSavedVariables();
	
	-- hook the chat handler
	ChatFrame_AddMessageEventFilter( "CHAT_MSG_WHISPER", HoJ_Filter_Whisper );

	HoJ:UpdateRole();
	
	HoJ:SetupAddOnOptions();

--[[-- debug.. show spells in spellbook
	for i=1,1000 do
	   local spellName, spellSubName = GetSpellBookItemName(i, BOOKTYPE_SPELL)
	   if spellName == nil then
		  break
	   end
	   -- use spellName and spellSubName here
	   DEFAULT_CHAT_FRAME:AddMessage( spellName .. '(' .. spellSubName .. ')' )
	end	
--]]
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_HoPo
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_HoPo( self ) 
	--HoJ:DebugMsg( "Set Up HoPo" );
	HoJ:EnableFrame( self );
	
	HoJ._HoPo[0] = self.HP1;
	HoJ._HoPoGlow[0] = self.HP1Glow;
	HoJ._HoPo[1] = self.HP2;
	HoJ._HoPoGlow[1] = self.HP2Glow;
	HoJ._HoPo[2] = self.HP3;
	HoJ._HoPoGlow[2] = self.HP3Glow;
	HoJ._HoPo[3] = self.HP4;
	HoJ._HoPoGlow[3] = self.HP4Glow;
	HoJ._HoPo[4] = self.HP5;
	HoJ._HoPoGlow[4] = self.HP5Glow;
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Buffs
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Buffs( self ) 
	--HoJ:DebugMsg( "Set Up Buffs" );
	HoJ:EnableFrame( self );

	HoJ._buffFrames = { 
		{
			frame = HoJ_Buff1Frame, 
			texture = HoJ_Buff1Texture,
			spellInfo = nil,
			text = HoJ_Buff1Text
		},
		{
			frame = HoJ_Buff2Frame, 
			texture = HoJ_Buff2Texture,
			spellInfo = nil,
			text = HoJ_Buff2Text
		},
		{
			frame = HoJ_Buff3Frame, 
			texture = HoJ_Buff3Texture,
			spellInfo = nil,
			text = HoJ_Buff3Text
		},
		{
			frame = HoJ_Buff4Frame, 
			texture = HoJ_Buff4Texture,
			spellInfo = nil,
			text = HoJ_Buff4Text
		},
		{
			frame = HoJ_Buff5Frame, 
			texture = HoJ_Buff5Texture,
			spellInfo = nil,
			text = HoJ_Buff5Text
		}
	}; 
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Cooldowns
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Cooldowns( self ) 
	--HoJ:DebugMsg( "Set Up Skills" );
	HoJ:EnableFrame( self );
	
	HoJ._cooldownFrames = { 
		{
			frame = HoJ_Cooldown1Frame, 
			texture = HoJ_Cooldown1Texture,
			spellInfo = nil,
			text = HoJ_Cooldown1Text
		},
		{
			frame = HoJ_Cooldown2Frame, 
			texture = HoJ_Cooldown2Texture,
			spellInfo = nil,
			text = HoJ_Cooldown2Text
		},
		{
			frame = HoJ_Cooldown3Frame, 
			texture = HoJ_Cooldown3Texture,
			spellInfo = nil,
			text = HoJ_Cooldown3Text
		},
		{
			frame = HoJ_Cooldown4Frame, 
			texture = HoJ_Cooldown4Texture,
			spellInfo = nil,
			text = HoJ_Cooldown4Text
		},
		{
			frame = HoJ_Cooldown5Frame, 
			texture = HoJ_Cooldown5Texture,
			spellInfo = nil,
			text = HoJ_Cooldown5Text
		},
		{
			frame = HoJ_Cooldown6Frame, 
			texture = HoJ_Cooldown6Texture,
			spellInfo = nil,
			text = HoJ_Cooldown6Text
		}
	};
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Rotation
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Rotation( self )
	HoJ:EnableFrame( self );

	HoJ._rotationFrames = {
		[1] = {
			frame = HoJ_RotationFrame, 
			texture = HoJ_RotationTexture,
			spellInfo = nil,
			text = HoJ_RotationText,
			cooldown = HoJ_RotationCooldown
		}
	};
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Inquisition
-- this is pre-WoD stuff
-- FIXME: remove the call to this from window template
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Inquisition( self ) 
	-- HoJ:DebugMsg( "Set Up Inquisition" );
	HoJ:EnableFrame( self );
	
	HoJ_InqTexture:SetTexture( HoJ.SpellInfo.inquisition.icon );
	HoJ_InqTexture:Show();
	local fontName, fontHeight, fontFlags = HoJ_InqTimer:GetFont();
	HoJ_InqTimer:SetFont( fontName, fontHeight, "THICKOUTLINE" );
	HoJ_InqTimer:SetTextColor( 1.0, 0.0, 0.0 );
	
	if ( HoJ._role ~= HOJ_ROLE_DPS ) then
		HoJ_InqFrameCooldown:SetAlpha( 0.0 );
		HoJ_InqTexture:SetAlpha( 0.0 );
		HoJ_InqTimer:SetAlpha( 0.0 );
	end
end
